#!/usr/bin/env python
# noinspection PyPep8Naming
"""To save memory, error/warning data is stored as a tuple:
    element 0: MessageTupleAdaptor instance (immutable, shared across all errors/
                    warnings of the same type)
    element 1: python memory address for the object or object reference
    element 2: rich object address (immutable and shared by all instances
                    of warnings for the same object)
    element 3: the "data" argument for the warning. must be hashable.
"""
from .warning_codes import NexsonWarningCodes


try:
    # noinspection PyCompatibility
    from cStringIO import StringIO
except ImportError:
    from io import StringIO
import codecs
import json

# monkey patching of NexsonWarningCodes causes lots of warnings
# pylint: disable=E1101
import logging
from peyutil import get_utf_8_string_io_writer
_LOG = logging.getLogger(__name__)


class MessageTupleAdaptor(object):
    """Base class for the Warning/Error types generated during validation.

    This base class provides the basic functionality of keeping
    track of the "address" of the element that triggered the warning,
    the severity code, and methods for writing to free text stream or JSON.
    """

    def write(self, err_tuple, s, prefix):
        raise NotImplementedError('MessageTupleAdaptor.write')

    def __init__(self):
        self.code = None

    def __unicode__(self, err_tuple, prefix=''):
        strio, wrapper = get_utf_8_string_io_writer()
        self.write(err_tuple, wrapper, prefix)
        return strio.getvalue()

    def getvalue(self, err_tuple, prefix=''):
        return self.__unicode__(err_tuple, prefix=prefix)

    def as_dict(self, err_tuple):
        addr = err_tuple[2]
        assert self.code is not None
        return {'@code': NexsonWarningCodes.facets[self.code],  # pylint: disable=E1126
                # 'comment': self.__unicode__(err_tuple),
                'data': self.convert_data_for_json(err_tuple),
                'refersTo': addr.path
                }

    def convert_data_for_json(self, err_tuple):  # pylint: disable=R0201
        return err_tuple[3]

    def _write_message_suffix(self, err_tuple, out):  # pylint: disable=R0201
        addr = err_tuple[2]
        addr.write_path_suffix_str(out)


# noinspection PyMissingConstructor
class _StrListDataWarningType(MessageTupleAdaptor):
    """Adaptor for warning with data being a list of strings
    """

    def __init__(self):
        self.code = None
        self.format = '{p}'

    def write(self, err_tuple, outstream, prefix):
        data = err_tuple[3]
        ds = '", "'.join(data)
        outstream.write(self.format.format(p=prefix, d=ds))
        self._write_message_suffix(err_tuple, outstream)


class _ArgumentlessWarningType(MessageTupleAdaptor):
    """Adaptor for warning with no data
    """

    def write(self, err_tuple, outstream, prefix):
        # noinspection PyUnresolvedReferences
        outstream.write(self.format.format(p=prefix))
        self._write_message_suffix(err_tuple, outstream)


class _StrArgumentWarningType(MessageTupleAdaptor):
    """Adaptor for warning with data being a string
    """

    def write(self, err_tuple, outstream, prefix):
        ds = err_tuple[3]
        # noinspection PyUnresolvedReferences
        outstream.write(self.format.format(p=prefix, d=ds))
        self._write_message_suffix(err_tuple, outstream)


# noinspection PyMissingConstructor
class RepeatedOTUWarningType(_StrListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.REPEATED_OTU
        self.format = '{p}OTU id found in mutliple nodes. id(s): "{d}"'


# noinspection PyMissingConstructor
class UnrecognizedKeyWarningType(_StrListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.UNRECOGNIZED_KEY
        self.format = '{p}Unrecognized key(s): "{d}"'


# noinspection PyMissingConstructor
class MissingMandatoryKeyWarningType(_StrListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.MISSING_MANDATORY_KEY
        self.format = '{p}MMultipleTipsToSameOttIdWarningissing required key(s): "{d}"'


# noinspection PyMissingConstructor
class MissingOptionalKeyWarningType(_StrListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.MISSING_OPTIONAL_KEY
        self.format = '{p}Missing optional key(s): "{d}"'


# noinspection PyMissingConstructor
class MissingExpectedListWarningType(_StrListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.MISSING_LIST_EXPECTED
        self.format = '{p}Expected a list for key(s): "{d}"'


# noinspection PyMissingConstructor
class MissingCrucialContentWarningType(_StrListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.MISSING_CRUCIAL_CONTENT
        self.format = '{p}Further validation amd acceptance requires more content; specifically keys(s): "{d}"'


# noinspection PyMissingConstructor
class MultipleRootsWarningType(_StrListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.MULTIPLE_ROOT_NODES
        self.format = '{p}Multiple nodes flagged as the root: "{d}"'


# noinspection PyMissingConstructor
class NoRootWarningType(_ArgumentlessWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.MULTIPLE_ROOT_NODES
        self.format = '{p}Multiple nodes flagged as the root:'


# noinspection PyMissingConstructor
class NodeWithMultipleParentsType(_StrListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.MULTIPLE_EDGES_FOR_NODES
        self.format = '{p} node ID is the target of multiple edges. Node ID(s): "{d}"'


# noinspection PyMissingConstructor
class TreeCycleWarningType(_StrListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.CYCLE_DETECTED
        self.format = '{p} node ID involved in a cycle in the tree. Node ID(s): "{d}"'


# noinspection PyMissingConstructor
class ReferencedIDNotFoundWarningType(_StrListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.REFERENCED_ID_NOT_FOUND
        self.format = '{p}An ID reference(s) did not match a previous ID: "{d}"'


# noinspection PyMissingConstructor
class RepeatedIDWarningType(_StrListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.REPEATED_ID
        self.format = '{p}Object identifier(s) repeated: "{d}"'


class _ObjListDataWarningType(_StrListDataWarningType):
    def convert_data_for_json(self, err_tuple):
        return [json.loads(i) for i in err_tuple[3]]


# noinspection PyMissingConstructor
class UnparseableMetaWarningType(_ObjListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.UNPARSEABLE_META
        self.format = '{p}meta(s) with out @property or @rel: "{d}"'


# noinspection PyMissingConstructor
class UnreachableNodeWarningType(_StrListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.UNREACHABLE_NODE
        self.format = '{p}Nodes not connected to the root of the tree: "{d}"'


# noinspection PyMissingConstructor
class InvalidKeyWarningType(_StrListDataWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.INVALID_PROPERTY_VALUE
        self.format = '{p}Invalid or inappropriate key: "{d}"'


# pylint: disable=R0921
class WrongValueTypeWarningType(MessageTupleAdaptor):  # pylint: disable=W0613
    def write(self, err_tuple, outstream, prefix):
        raise NotImplementedError('WrongValueTypeWarningType.write')

    def convert_data_for_json(self, err_tuple):
        rl = []
        for k, v, et in err_tuple[3]:
            rl.append({'key': k, 'type': v, 'expected': et})
        return rl

    def __init__(self):
        MessageTupleAdaptor.__init__(self)
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.INCORRECT_VALUE_TYPE
        self.format = '{p}value for key not the expected type: "{d}"'


class MultipleTipsToSameOttIdWarningType(MessageTupleAdaptor):  # pylint: disable=R0921, W0613
    def __init__(self):
        MessageTupleAdaptor.__init__(self)
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.MULTIPLE_TIPS_MAPPED_TO_OTT_ID
        self.format = '{p}Multiple otus mapping to the same OTT ID used in the same tree: "{d}"'

    def write(self, err_tuple, outstream, prefix):
        raise NotImplementedError('MultipleTipsToSameOttIdWarningType.write')

    def convert_data_for_json(self, err_tuple):
        return [i for i in err_tuple[3]]


# noinspection PyMissingConstructor
class MaxSizeExceededWarningType(_StrArgumentWarningType):
    def __init__(self):
        # noinspection PyUnresolvedReferences
        self.code = NexsonWarningCodes.MAX_SIZE_EXCEEDED
        self.format = '{p}Maximum size exceeded: {d}'


# A single, immutable, global instance of each warning type is created

InvalidKeyWarning = InvalidKeyWarningType()
MissingCrucialContentWarning = MissingCrucialContentWarningType()
MissingExpectedListWarning = MissingExpectedListWarningType()
MissingMandatoryKeyWarning = MissingMandatoryKeyWarningType()
MissingOptionalKeyWarning = MissingOptionalKeyWarningType()
MultipleRootsWarning = MultipleRootsWarningType()
MultipleTipsToSameOttIdWarning = MultipleTipsToSameOttIdWarningType()
NodeWithMultipleParents = NodeWithMultipleParentsType()
NoRootWarning = NoRootWarningType()
ReferencedIDNotFoundWarning = ReferencedIDNotFoundWarningType()
RepeatedIDWarning = RepeatedIDWarningType()
RepeatedOTUWarning = RepeatedOTUWarningType()
TreeCycleWarning = TreeCycleWarningType()
UnparseableMetaWarning = UnparseableMetaWarningType()
UnreachableNodeWarning = UnreachableNodeWarningType()
UnrecognizedKeyWarning = UnrecognizedKeyWarningType()
WrongValueTypeWarning = WrongValueTypeWarningType()
MaxSizeExceededWarning = MaxSizeExceededWarningType()



def gen_InvalidKeyWarning(addr, pyid, logger, severity, **kwargs):
    _key_list_warning(InvalidKeyWarning, kwargs['key_list'], addr, pyid, logger, severity)


def gen_MissingCrucialContentWarning(addr, pyid, logger, severity, **kwargs):
    _key_list_warning(MissingCrucialContentWarning, kwargs['key_list'], addr, pyid, logger, severity)


def gen_MissingExpectedListWarning(addr, pyid, logger, severity, **kwargs):
    _key_list_warning(MissingExpectedListWarning, kwargs['key_list'], addr, pyid, logger, severity)


def gen_MissingMandatoryKeyWarning(addr, pyid, logger, severity, **kwargs):
    _key_list_warning(MissingMandatoryKeyWarning, kwargs['key_list'], addr, pyid, logger, severity)


def gen_MissingOptionalKeyWarning(addr, pyid, logger, severity, **kwargs):
    _key_list_warning(MissingOptionalKeyWarning, kwargs['key_list'], addr, pyid, logger, severity)


def gen_MultipleRootsWarning(addr, pyid, logger, severity, **kwargs):
    _key_list_warning(MultipleRootsWarning, kwargs['node_id_list'], addr, pyid, logger, severity)


def gen_MultipleTipsToSameOttIdWarning(addr, pyid, logger, severity, **kwargs):
    k_list = kwargs['otu_sets']
    k_list.sort()
    k_list = tuple(k_list)
    t = (MultipleTipsToSameOttIdWarning, pyid, addr, k_list)
    logger.register_new_messages(t, severity=severity)


def gen_NodeWithMultipleParents(addr, pyid, logger, severity, **kwargs):
    _key_list_warning(NodeWithMultipleParents, kwargs['node_id_list'], addr, pyid, logger, severity)


def gen_TreeCycleWarning(addr, pyid, logger, severity, **kwargs):
    _key_list_warning(TreeCycleWarning, kwargs['node_id_list'], addr, pyid, logger, severity)


# noinspection PyUnusedLocal
def gen_NoRootWarning(addr, pyid, logger, severity, **kwargs):  # pylint: disable=W0613
    _argumentless_warning(NoRootWarning, addr, pyid, logger, severity)


def gen_ReferencedIDNotFoundWarning(addr, pyid, logger, severity, **kwargs):
    _key_list_warning(ReferencedIDNotFoundWarning, kwargs['key_list'], addr, pyid, logger, severity)


def gen_RepeatedIDWarning(addr, pyid, logger, severity, **kwargs):
    _key_list_warning(RepeatedIDWarning, kwargs['key_list'], addr, pyid, logger, severity)


def gen_RepeatedOTUWarning(addr, pyid, logger, severity, **kwargs):
    _key_list_warning(RepeatedOTUWarning, kwargs['key_list'], addr, pyid, logger, severity)


def gen_UnparseableMetaWarning(addr, pyid, logger, severity, **kwargs):
    _obj_list_warning(UnparseableMetaWarning, kwargs['obj_list'], addr, pyid, logger, severity)


def gen_UnreachableNodeWarning(addr, pyid, logger, severity, **kwargs):
    _key_list_warning(UnreachableNodeWarning, kwargs['key_list'], addr, pyid, logger, severity)


def gen_UnrecognizedKeyWarning(addr, pyid, logger, severity, **kwargs):
    _key_list_warning(UnrecognizedKeyWarning, kwargs['key_list'], addr, pyid, logger, severity)


def gen_WrongValueTypeWarning(addr, pyid, logger, severity, **kwargs):
    key_val_type_list = tuple([(k, type(v).__name__, t) for k, v, t in kwargs['key_val_type_list']])
    t = (WrongValueTypeWarning, pyid, addr, key_val_type_list)
    logger.register_new_messages(t, severity=severity)


def gen_MaxSizeExceededWarning(addr, pyid, logger, severity, **kwargs):
    _str_argument_warning(MaxSizeExceededWarning, addr, pyid, logger, severity, kwargs['message'])


# factory functions that call register_new_messages
def _obj_list_warning(wt, k_list, addr, pyid, logger, severity):
    k_list = tuple([json.dumps(i) for i in k_list])
    t = (wt, pyid, addr, k_list)
    logger.register_new_messages(t, severity=severity)


def _key_list_warning(wt, k_list, addr, pyid, logger, severity):
    k_list.sort()
    k_list = tuple(k_list)
    t = (wt, pyid, addr, k_list)
    # _LOG.debug("t=" + str(t))
    logger.register_new_messages(t, severity=severity)


def _argumentless_warning(wt, addr, pyid, logger, severity):
    t = (wt, pyid, addr, None)
    logger.register_new_messages(t, severity=severity)


def _str_argument_warning(wt, addr, pyid, logger, severity, message):
    t = (wt, pyid, addr, message)
    logger.register_new_messages(t, severity=severity)


# noinspection PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences
factory2code = {gen_MultipleTipsToSameOttIdWarning: NexsonWarningCodes.MULTIPLE_TIPS_MAPPED_TO_OTT_ID,
                gen_NoRootWarning: NexsonWarningCodes.MULTIPLE_ROOT_NODES,
                gen_UnrecognizedKeyWarning: NexsonWarningCodes.UNRECOGNIZED_KEY,
                gen_MissingOptionalKeyWarning: NexsonWarningCodes.MISSING_OPTIONAL_KEY,
                gen_MissingExpectedListWarning: NexsonWarningCodes.MISSING_LIST_EXPECTED,
                gen_UnparseableMetaWarning: NexsonWarningCodes.UNPARSEABLE_META,
                gen_ReferencedIDNotFoundWarning: NexsonWarningCodes.REFERENCED_ID_NOT_FOUND,
                gen_TreeCycleWarning: NexsonWarningCodes.CYCLE_DETECTED,
                gen_RepeatedIDWarning: NexsonWarningCodes.REPEATED_ID,
                gen_MissingCrucialContentWarning: NexsonWarningCodes.MISSING_CRUCIAL_CONTENT,
                gen_UnreachableNodeWarning: NexsonWarningCodes.UNREACHABLE_NODE,
                gen_RepeatedOTUWarning: NexsonWarningCodes.REPEATED_OTU,
                gen_NodeWithMultipleParents: NexsonWarningCodes.MULTIPLE_EDGES_FOR_NODES,
                gen_WrongValueTypeWarning: NexsonWarningCodes.INCORRECT_VALUE_TYPE,
                gen_MissingMandatoryKeyWarning: NexsonWarningCodes.MISSING_MANDATORY_KEY,
                gen_MultipleRootsWarning: NexsonWarningCodes.MULTIPLE_ROOT_NODES,
                gen_InvalidKeyWarning: NexsonWarningCodes.INVALID_PROPERTY_VALUE,
                gen_MaxSizeExceededWarning: NexsonWarningCodes.MAX_SIZE_EXCEEDED,
                }



#!/usr/bin/env python
"""Direct2OptimalNexson class"""
from .helper import (NexsonConverter, get_nexml_el, _get_index_list_of_values, 
                     _index_list_of_values, BY_ID_HONEY_BADGERFISH, NexsonError)
import logging

_LOG = logging.getLogger(__name__)


class Direct2OptimalNexson(NexsonConverter):
    """Conversion of the direct port of NeXML to JSON (v 1.0)
    to the more optimized version (v 1.2).
    This is a dict-to-dict in-place conversion. No serialization is included.
    """

    def __init__(self, conv_cfg):
        NexsonConverter.__init__(self, conv_cfg)
        self.suppress_label_if_ott_taxon = getattr(conv_cfg, 'suppress_label_if_ott_taxon', True)

    def convert_otus(self, otus_list):
        otusById = dict((i['@id'], i) for i in otus_list)
        otusElementOrder = [i['@id'] for i in otus_list]
        otusIdToOtuObj = {}
        for oid, otus_el in otusById.items():
            o_list = _index_list_of_values(otus_el, 'otu')
            otuById = dict((i['@id'], i) for i in o_list)
            otusIdToOtuObj[oid] = otuById
        # If all that succeeds, add the new object to the dict, creating a fat structure
        for k, v in otusIdToOtuObj.items():
            otusById[k]['otuById'] = v
        # Make the struct leaner
        if self.remove_old_structs:
            for v in otusById.values():
                del v['@id']
            for k, otu_obj in otusIdToOtuObj.items():
                o = otusById[k]
                del o['otu']
                for v in otu_obj.values():
                    del v['@id']
                    # move @label to ^ot:manualLabel if it is not ottTaxonName
                    # if self.suppress_label_if_ott_taxon:
                    #     if '@label' in v:
                    #         tax_name = v.get('^ot:ottTaxonName')
                    #         label = v.get('@label').strip()
                    #         if tax_name is None:
                    #             orig_name = v.get('^ot:originalLabel')
                    #             assert(orig_name is not None)
                    #             if label != orig_name:
                    #                 v['^ot:manualLabel'] = label
                    #         elif label != tax_name:
                    #             v['^ot:manualLabel'] = label
                    #         del v['@label']
        return otusById, otusElementOrder

    def convert_tree(self, tree):
        """Return (tree_id, tree) or None (if the tree has no edges).
        """
        nodeById = {}
        root_node = None
        node_list = _index_list_of_values(tree, 'node')
        for node in node_list:
            nodeById[node['@id']] = node
            r = node.get('@root')
            # _LOG.debug(' node {} @root={}'.format(node['@id'], r))
            if r in [True, 'true']:  # @TEMP accepting true or "true"
                assert root_node is None
                root_node = node
        assert root_node is not None
        edgeBySourceId = {}
        edge_list = _get_index_list_of_values(tree, 'edge')
        for edge in edge_list:
            sourceId = edge['@source']
            eid = edge['@id']
            del edge['@id']
            byso = edgeBySourceId.setdefault(sourceId, {})
            byso[eid] = edge
        # If all that succeeds, add the new object to the dict, creating a fat structure
        tree['nodeById'] = nodeById
        tree['edgeBySourceId'] = edgeBySourceId
        tree['^ot:rootNodeId'] = root_node['@id']
        # Make the struct leaner
        tid = tree['@id']
        if self.remove_old_structs:
            del tree['@id']
            del tree['node']
            try:
                del tree['edge']
            except:
                # Tree Tr75035 in http://treebase.org/treebase-web/search/study/summary.html?id=14763
                #   is empty. in NeXML that shows up as a tree with a node but no edges.
                #   See https://github.com/OpenTreeOfLife/opentree/issues/641
                # TODO: returning None seems safest, but could cull trees with just metadata.
                #       but creating a fake tree for metadata is ugly. So, I'm fine with not
                #       supporting this.
                _LOG.warning('Tree with ID "{}" is being dropped because it has no edges'.format(tid))
                assert not edge_list
                return None
            for node in node_list:
                if '^ot:isLeaf' in node:
                    del node['^ot:isLeaf']
                del node['@id']
        return tid, tree

    def convert(self, obj):
        """Takes a dict corresponding to the honeybadgerfish JSON blob of the 1.0.* type and
        converts it to BY_ID_HONEY_BADGERFISH version. The object is modified in place
        and returned.
        """
        if self.pristine_if_invalid:
            raise NotImplementedError('pristine_if_invalid option is not supported yet')

        nex = get_nexml_el(obj)
        assert nex
        # Create the new objects as locals. This section should not
        #   mutate obj, so that if there is an exception the object
        #   is unchanged on the error exit
        otus = _index_list_of_values(nex, 'otus')
        o_t = self.convert_otus(otus)
        otusById, otusElementOrder = o_t
        trees = _get_index_list_of_values(nex, 'trees')
        treesById = dict((i['@id'], i) for i in trees)
        treesElementOrder = [i['@id'] for i in trees]
        if len(treesById) != len(treesElementOrder):
            trees_id_set = set()
            for tgid in treesElementOrder:
                if tgid in trees_id_set:
                    raise NexsonError('Repeated trees element id "{}"'.format(tgid))
                trees_id_set.add(tgid)
        tree_id_set = set()
        treeContainingObjByTreesId = {}
        for tree_group in trees:
            # _LOG.debug('converting tree group {} to by_id'.format(tree_group['@id']))
            treeById = {}
            treeElementOrder = []
            tree_array = _get_index_list_of_values(tree_group, 'tree')
            for tree in tree_array:
                # _LOG.debug('# pre-convert keys = {}'.format(tree.keys()))
                t_t = self.convert_tree(tree)
                if t_t is None:
                    continue
                tid, tree_alias = t_t  # pylint: disable=W0633
                if tid in tree_id_set:
                    raise NexsonError('Repeated tree element id "{}"'.format(tid))
                tree_id_set.add(tid)

                # _LOG.debug('converting tree {} to by_id'.format(tid))
                # _LOG.debug('# post-convert keys = {}'.format(tree.keys()))
                assert tree_alias is tree
                treeById[tid] = tree
                treeElementOrder.append(tid)
            treeContainingObjByTreesId[tree_group['@id']] = treeById
            tree_group['^ot:treeElementOrder'] = treeElementOrder

        # If all that succeeds, add the new object to the dict, creating a fat structure
        nex['otusById'] = otusById
        nex['^ot:otusElementOrder'] = otusElementOrder
        nex['treesById'] = treesById
        nex['^ot:treesElementOrder'] = treesElementOrder
        for k, v in treeContainingObjByTreesId.items():
            treesById[k]['treeById'] = v
        nex['@nexml2json'] = str(BY_ID_HONEY_BADGERFISH)
        # Make the struct leaner
        if self.remove_old_structs:
            del nex['otus']
            del nex['trees']
            for k, v in treesById.items():
                if 'tree' in v:
                    del v['tree']
                del v['@id']
        return obj



#!/usr/bin/env python
"""Optimal2DirectNexson class"""
from .helper import (NexsonConverter,
                     get_nexml_el,
                     DIRECT_HONEY_BADGERFISH,
                     NexsonError)
import logging

_LOG = logging.getLogger(__name__)


# noinspection PyPep8Naming
class Optimal2DirectNexson(NexsonConverter):
    """Conversion of the optimized (v 1.2) version of NexSON to
    the more direct (v 1.0) port of NeXML
    This is a dict-to-dict in-place conversion. No serialization is included.
    """

    def __init__(self, conv_cfg):
        NexsonConverter.__init__(self, conv_cfg)

    def convert_otus(self, otusById, otusElementOrder):
        if self.pristine_if_invalid:
            raise NotImplementedError('pristine_if_invalid option is not supported yet')
        otu_group_list = []
        for oid in otusElementOrder:
            otu_group = otusById[oid]
            otu_group['@id'] = oid
            otu_list = []
            otu_by_id = otu_group['otuById']
            otu_id_list = list(otu_by_id.keys())
            otu_id_list.sort()  # not necessary, but will give us a consistent order...
            for otu_id in otu_id_list:
                otu = otu_by_id[otu_id]
                otu['@id'] = otu_id
                otu_list.append(otu)
            otu_group['otu'] = otu_list
            if self.remove_old_structs:
                del otu_group['otuById']
            otu_group_list.append(otu_group)
        return otu_group_list

    def convert_tree(self, tree):
        if self.pristine_if_invalid:
            raise NotImplementedError('pristine_if_invalid option is not supported yet')
        nodeById = tree['nodeById']
        edgeBySourceId = tree['edgeBySourceId']
        root_node_id = tree['^ot:rootNodeId']
        node_list = []
        edge_list = []
        curr_node_id = root_node_id
        edge_stack = []
        node_set_written = set()
        edge_set_written = set()
        while True:
            curr_node = nodeById[curr_node_id]
            curr_node['@id'] = curr_node_id
            assert curr_node_id not in node_set_written
            node_set_written.add(curr_node_id)
            node_list.append(curr_node)
            sub_edge_dict = edgeBySourceId.get(curr_node_id)
            if sub_edge_dict:
                ks = list(sub_edge_dict.keys())
                ks.sort()
                sub_edge_list = [(ski, sub_edge_dict[ski]) for ski in ks]
                eid, edge = sub_edge_list[0]
                to_stack = sub_edge_list[-1:0:-1]
                edge_stack.extend(to_stack)
            else:
                curr_node['^ot:isLeaf'] = True
                if not edge_stack:
                    break
                eid, edge = edge_stack.pop(-1)
            edge['@id'] = eid
            edge_list.append(edge)
            assert eid not in edge_set_written
            edge_set_written.add(eid)
            curr_node_id = edge['@target']
        for n in nodeById.values():
            assert n['@id'] in node_set_written
        tree['node'] = node_list
        tree['edge'] = edge_list
        if self.remove_old_structs:
            del tree['nodeById']
            del tree['edgeBySourceId']
            del tree['^ot:rootNodeId']
        return tree

    def convert_trees(self, treesById, treesElementOrder):

        if self.pristine_if_invalid:
            raise NotImplementedError('pristine_if_invalid option is not supported yet')
        trees_group_list = []
        tree_id_set = set()
        trees_id_set = set()
        for tgid in treesElementOrder:
            # _LOG.debug('tgid = ' + tgid)
            tree_group = treesById[tgid]
            if tgid in trees_id_set:
                raise NexsonError('Repeated trees element id "{}"'.format(tgid))
            trees_id_set.add(tgid)
            tree_group['@id'] = tgid
            treeElementOrder = tree_group['^ot:treeElementOrder']
            tree_list = []
            tree_by_id = tree_group['treeById']
            for tree_id in treeElementOrder:
                if tree_id in tree_id_set:
                    raise NexsonError('Repeated tree element id "{}"'.format(tree_id))
                tree_id_set.add(tree_id)
                tree = tree_by_id[tree_id]
                # _LOG.debug('pre-convert  tree(id={}).keys = {}'.format(tree_id, tree.keys()))
                self.convert_tree(tree)
                # _LOG.debug('post-convert tree(id={}).keys = {}'.format(tree_id, tree.keys()))
                tree['@id'] = tree_id
                tree_list.append(tree)
            tree_group['tree'] = tree_list
            if self.remove_old_structs:
                del tree_group['treeById']
                del tree_group['^ot:treeElementOrder']
            trees_group_list.append(tree_group)
        return trees_group_list

    def convert(self, obj):
        """Takes a dict corresponding to the honeybadgerfish JSON blob of the 1.2.* type and
        converts it to DIRECT_HONEY_BADGERFISH version. The object is modified in place
        and returned.
        """
        if self.pristine_if_invalid:
            raise NotImplementedError('pristine_if_invalid option is not supported yet')

        nex = get_nexml_el(obj)
        assert nex
        # Create the new objects as locals. This section should not
        #   mutate obj, so that if there is an exception the object
        #   is unchanged on the error exit
        otusById = nex['otusById']
        otusElementOrder = nex['^ot:otusElementOrder']
        otus = self.convert_otus(otusById, otusElementOrder)
        nex['otus'] = otus
        treesById = nex['treesById']
        treesElementOrder = nex['^ot:treesElementOrder']
        trees = self.convert_trees(treesById, treesElementOrder)
        # add the locals to the object
        nex['trees'] = trees
        nex['@nexml2json'] = str(DIRECT_HONEY_BADGERFISH)
        # Make the struct leaner
        if self.remove_old_structs:
            del nex['otusById']
            del nex['^ot:otusElementOrder']
            del nex['treesById']
            del nex['^ot:treesElementOrder']
        return obj



#!/usr/bin/env python
"""Nexson2Nexml class"""
from .helper import (NexsonConverter,
                     _add_value_to_dict_bf,
                     _convert_hbf_meta_val_for_xml,
                     _index_list_of_values,
                     _is_badgerfish_version)
from peyutil import UNICODE
import xml.dom.minidom
import logging

_LOG = logging.getLogger(__name__)


def _create_sub_el(doc, parent, tag, attrib, data=None):
    """Creates and xml element for the `doc` with the given `parent`
    and `tag` as the tagName.
    `attrib` should be a dictionary of string keys to primitives or dicts
        if the value is a dict, then the keys of the dict are joined with
        the `attrib` key using a colon. This deals with the badgerfish
        convention of nesting xmlns: attributes in a @xmnls object
    If `data` is not None, then it will be written as data. If it is a boolean,
        the xml true false will be writtten. Otherwise it will be
        converted to python unicode string, stripped and written.
    Returns the element created
    """
    el = doc.createElement(tag)
    if attrib:
        if ('id' in attrib) and ('about' not in attrib):
            about_val = '#' + attrib['id']
            el.setAttribute('about', about_val)
        for att_key, att_value in attrib.items():
            if isinstance(att_value, dict):
                for inner_key, inner_val in att_value.items():
                    rk = ':'.join([att_key, inner_key])
                    el.setAttribute(rk, inner_val)
            else:
                el.setAttribute(att_key, att_value)
    if parent:
        parent.appendChild(el)
    if data is not None:
        if data is True:
            el.appendChild(doc.createTextNode('true'))
        elif data is False:
            el.appendChild(doc.createTextNode('false'))
        else:
            u = UNICODE(data).strip()
            if u:
                el.appendChild(doc.createTextNode(u))
    return el


def _convert_bf_meta_val_for_xml(blob):
    if not isinstance(blob, list):
        blob = [blob]
    first_blob = blob[0]
    try:
        try:
            if first_blob.get("@xsi:type") == "nex:LiteralMeta":
                return first_blob["@property"], blob
        except:
            pass
        return first_blob["@rel"], blob
    except:
        return "", blob


class Nexson2Nexml(NexsonConverter):
    """Conversion of the optimized (v 1.2) version of NexSON to
    the more direct (v 1.0) port of NeXML
    This is a dict-to-minidom-doc conversion. No serialization is included.
    """

    def __init__(self, conv_cfg):
        NexsonConverter.__init__(self, conv_cfg)
        self.input_format = conv_cfg.input_format
        self.use_default_root_atts = conv_cfg.get('use_default_root_atts', True)
        self.otu_label = conv_cfg.get('otu_label', 'ot:originalLabel')
        if self.otu_label.startswith('^'):
            self.otu_label = self.otu_label[1:]
        self._migrating_from_bf = _is_badgerfish_version(self.input_format)
        # TreeBase and phylografter trees often lack the tree xsi:type
        self._adding_tree_xsi_type = True
        # we have started using ot:ottTaxonName, ot:originalLabel or ot:ottId
        self._creating_otu_label = True

    def convert(self, blob):
        doc = xml.dom.minidom.Document()
        converted_root_el = False
        if 'nexml' in blob:
            converted_root_el = True
            blob['nex:nexml'] = blob['nexml']
            del blob['nexml']
        self._top_level_build_xml(doc, blob)
        if converted_root_el:
            blob['nexml'] = blob['nex:nexml']
            del blob['nex:nexml']

        return doc

    def _partition_keys_for_xml(self, o):
        """Breaks o into four content type by key syntax:
            attrib keys (start with '@'),
            text (value associated with the '$' or None),
            child element keys (all others)
            meta element
        """
        ak = {}
        tk = None
        ck = {}
        mc = {}
        # _LOG.debug('o = {o}'.format(o=o))
        for k, v in o.items():
            if k.startswith('@'):
                if k == '@xmlns':
                    if '$' in v:
                        ak['xmlns'] = v['$']
                    for nsk, nsv in v.items():
                        if nsk != '$':
                            ak['xmlns:' + nsk] = nsv
                else:
                    s = k[1:]
                    if isinstance(v, bool):
                        v = u'true' if v else u'false'
                    ak[s] = UNICODE(v)
            elif k == '$':
                tk = v
            elif k.startswith('^') and (not self._migrating_from_bf):
                s = k[1:]
                val = _convert_hbf_meta_val_for_xml(s, v)
                _add_value_to_dict_bf(mc, s, val)
            elif (k == u'meta') and self._migrating_from_bf:
                s, val = _convert_bf_meta_val_for_xml(v)
                _add_value_to_dict_bf(mc, s, val)
            else:
                ck[k] = v
        return ak, tk, ck, mc

    def _top_level_build_xml(self, doc, obj_dict):
        if self.use_default_root_atts:
            root_atts = {
                "xmlns:nex": "http://www.nexml.org/2009",
                "xmlns": "http://www.nexml.org/2009",
                "xmlns:xsd": "http://www.w3.org/2001/XMLSchema#",
                "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
                "xmlns:ot": "http://purl.org/opentree/nexson",
            }
        else:
            root_atts = {}
        # extra = {
        #     "xmlns:dc": "http://purl.org/dc/elements/1.1/",
        #     "xmlns:dcterms": "http://purl.org/dc/terms/",
        #     "xmlns:prism": "http://prismstandard.org/namespaces/1.2/basic/",
        #     "xmlns:rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        #     "xmlns:rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        #     "xmlns:skos": "http://www.w3.org/2004/02/skos/core#",
        #     "xmlns:tb": "http://purl.org/phylo/treebase/2.0/terms#",
        # }
        base_keys = list(obj_dict.keys())
        assert len(base_keys) == 1
        root_name = base_keys[0]
        root_obj = obj_dict[root_name]
        atts, data, children, meta_children = self._partition_keys_for_xml(root_obj)
        if 'generator' not in atts:
            atts['generator'] = 'org.opentreeoflife.api.nexsonvalidator.nexson_nexml'
        if 'version' not in atts:
            atts['version'] = '0.9'
        if root_atts:
            for k, v in root_atts.items():
                atts[k] = v
        if ('id' in atts) and ('about' not in atts):
            atts['about'] = '#' + atts['id']
        if 'nexml2json' in atts:
            del atts['nexml2json']
        r = _create_sub_el(doc, doc, root_name, atts, data)
        self._add_meta_dict_to_xml(doc, r, meta_children)
        nexml_key_order = (('meta', None),
                           ('otus', (('meta', None),
                                     ('otu', None)
                                     )
                            ),
                           ('characters', (('meta', None),
                                           ('format', (('meta', None),
                                                       ('states', (('state', None),
                                                                   ('uncertain_state_set', None),
                                                                   )
                                                        ),
                                                       ('char', None)
                                                       ),
                                            ),
                                           ('matrix', (('meta', None),
                                                       ('row', None),
                                                       )
                                            ),
                                           ),
                            ),
                           ('trees', (('meta', None),
                                      ('tree', (('meta', None),
                                                ('node', None),
                                                ('edge', None)
                                                )
                                       )
                                      )
                            )
                           )
        self._add_dict_of_subtree_to_xml_doc(doc, r, children, nexml_key_order)

    def _add_subtree_list_to_xml_doc(self, doc, par, ch_list, key, key_order):
        for child in ch_list:
            if isinstance(child, dict):
                self._add_subtree_to_xml_doc(doc, par, child, key, key_order)
            else:
                ca = {}
                cc = {}
                mc = {}
                if isinstance(child, list) or isinstance(child, tuple) or isinstance(child, set):
                    for sc in child:
                        if isinstance(sc, dict):
                            self._add_subtree_to_xml_doc(doc, par, sc, key, key_order)
                        else:
                            cd = sc
                            cel = _create_sub_el(doc, par, key, ca, cd)
                            self._add_meta_dict_to_xml(doc, cel, mc)
                            self._add_dict_of_subtree_to_xml_doc(doc, cel, cc, key_order=None)
                else:
                    cd = child
                    cel = _create_sub_el(doc, par, key, ca, cd)
                    self._add_meta_dict_to_xml(doc, cel, mc)
                    self._add_dict_of_subtree_to_xml_doc(doc, cel, cc, key_order=None)

    def _add_dict_of_subtree_to_xml_doc(self,
                                        doc,
                                        parent,
                                        children_dict,
                                        key_order=None):
        written = set()
        if key_order:
            for t in key_order:
                k, nko = t
                assert nko is None or isinstance(nko, tuple)
                if k in children_dict:
                    chl = _index_list_of_values(children_dict, k)
                    written.add(k)
                    self._add_subtree_list_to_xml_doc(doc, parent, chl, k, nko)
        ksl = list(children_dict.keys())
        ksl.sort()
        for k in ksl:
            chl = _index_list_of_values(children_dict, k)
            if k not in written:
                self._add_subtree_list_to_xml_doc(doc, parent, chl, k, None)

    def _add_subtree_to_xml_doc(self,
                                doc,
                                parent,
                                subtree,
                                key,
                                key_order,
                                extra_atts=None,
                                del_atts=None):
        ca, cd, cc, mc = self._partition_keys_for_xml(subtree)
        if extra_atts is not None:
            ca.update(extra_atts)
        if del_atts is not None:
            for da in del_atts:
                if da in ca:
                    del ca[da]
        if self._adding_tree_xsi_type:
            if (key == 'tree') and (parent.tagName == 'trees') and ('xsi:type' not in ca):
                ca['xsi:type'] = 'nex:FloatTree'
        if self._creating_otu_label and (key == 'otu') and (parent.tagName == 'otus'):
            key_to_promote = self.otu_label  # need to verify that we are converting from 1.0 not 0.0..
            # _LOG.debug(str((key_to_promote, mc.keys())))
            if key_to_promote in mc:
                val = mc[key_to_promote]
                if isinstance(val, dict):
                    val = val['$']
                ca['label'] = str(val)
            elif key_to_promote in ca:
                ca['label'] = str(ca[key_to_promote])
        cel = _create_sub_el(doc, parent, key, ca, cd)
        self._add_meta_dict_to_xml(doc, cel, mc)
        self._add_dict_of_subtree_to_xml_doc(doc, cel, cc, key_order)
        return cel

    def _add_meta_dict_to_xml(self, doc, parent, meta_dict):
        """
        Values in the meta element dict are converted to a BadgerFish-style
            encoding (see _convert_hbf_meta_val_for_xml), so regardless of input_format,
            we treat them as if they were BadgerFish.
        """
        if not meta_dict:
            return
        key_list = list(meta_dict.keys())
        key_list.sort()
        for key in key_list:
            el_list = _index_list_of_values(meta_dict, key)
            for el in el_list:
                self._add_meta_value_to_xml_doc(doc, parent, el)

    def _add_meta_value_to_xml_doc(self, doc, parent, obj):
        """Values in the meta element dict are converted to a BadgerFish-style
            encoding (see _convert_hbf_meta_val_for_xml), so regardless of input_format,
            we treat them as if they were BadgerFish.
        """
        return self._add_subtree_to_xml_doc(doc,
                                            parent,
                                            subtree=obj,
                                            key='meta',
                                            key_order=None)



#!/usr/bin/env python
"""Direct2BadgerfishNexson class"""
from .helper import (NexsonConverter,
                     get_nexml_el,
                     _add_redundant_about,
                     _convert_hbf_meta_val_for_xml,
                     BADGER_FISH_NEXSON_VERSION)
import logging

_LOG = logging.getLogger(__name__)


# noinspection PyTypeChecker
class Direct2BadgerfishNexson(NexsonConverter):
    """Conversion of the direct form of honeybadgerfish
    HoneyBadgerFish (v 1.0) to "raw" Badgerfish + phylografter tweaks
    This is a dict-to-dict in-place conversion. No serialization is included.
    """

    def __init__(self, conv_cfg):
        NexsonConverter.__init__(self, conv_cfg)
        self.remove_old_structs = getattr(conv_cfg, 'remove_old_structs', True)

    def _recursive_convert_list(self, obj):
        for el in obj:
            if isinstance(el, dict):
                self._recursive_convert_dict(el)

    def _recursive_convert_dict(self, obj):
        _add_redundant_about(obj)  # rule 10...
        meta_list = []
        to_del = set()
        for k, v in obj.items():
            if k.startswith('^'):
                to_del.add(k)
                converted = _convert_hbf_meta_val_for_xml(k[1:], v)
                if isinstance(converted, list):
                    meta_list.extend(converted)
                else:
                    meta_list.append(converted)
            if isinstance(v, dict):
                self._recursive_convert_dict(v)
            elif isinstance(v, list):
                self._recursive_convert_list(v)
        for k in to_del:
            del obj[k]
        if meta_list:
            m = obj.get('meta')
            if m is None:
                obj['meta'] = meta_list
            elif isinstance(m, list):
                m.extend(meta_list)
            else:
                obj['meta'] = [m] + meta_list

    def _single_el_list_to_dicts(self, obj, tag, child_tag=None, grand_child_tag=None):
        el = obj.get(tag)
        if el:
            if isinstance(el, list):
                if len(el) == 1:
                    as_list = [el]
                    if child_tag is None:
                        obj[tag] = el[0]
                        return
                else:
                    as_list = el
            else:
                as_list = [el]
            if child_tag is not None:
                for sub in as_list:
                    self._single_el_list_to_dicts(sub, tag=child_tag, child_tag=grand_child_tag)

    def convert(self, obj):
        """Takes a dict corresponding to the honeybadgerfish JSON blob of the 1.0.* type and
        converts it to BY_ID_HONEY_BADGERFISH version. The object is modified in place
        and returned.
        """
        if self.pristine_if_invalid:
            raise NotImplementedError('pristine_if_invalid option is not supported yet')

        nex = get_nexml_el(obj)
        assert nex
        self._recursive_convert_dict(nex)
        nex['@nexml2json'] = str(BADGER_FISH_NEXSON_VERSION)
        self._single_el_list_to_dicts(nex, 'otus')
        self._single_el_list_to_dicts(nex, 'trees')
        #
        # otu and tree are always arrays in phylografter
        emulate_phylografter_pluralization = True
        if not emulate_phylografter_pluralization:
            self._single_el_list_to_dicts(nex, 'otus', 'otu')
            self._single_el_list_to_dicts(nex, 'trees', 'tree')
            self._single_el_list_to_dicts(nex, 'trees', 'tree', 'node')
            self._single_el_list_to_dicts(nex, 'trees', 'tree', 'edge')
        return obj



#!/usr/bin/env python
"""Classes for recording warnings and errors
"""
from ..syntax.helper import _add_value_to_dict_bf
from .helper import SeverityCodes, VERSION
from .warning_codes import NexsonWarningCodes
from functools import cmp_to_key
import logging
import platform
import sys

_LOG = logging.getLogger(__name__)


def _err_warn_summary(w):
    d = {}
    for el in w:
        msg_adapt_inst = el[0]
        r = msg_adapt_inst.as_dict(el)
        key = r['@code']
        del r['@code']
        _add_value_to_dict_bf(d, key, r)
    return d

# noinspection PyUnusedLocal
def _create_message_list(key, w, severity):  # pylint: disable=W0613
    d = []
    for el in w:
        msg_adapt_inst = el[0]
        r = msg_adapt_inst.as_dict(el)
        r['@severity'] = severity
        d.append(r)
    return d


_LIST_0 = [0]
_LIST_1 = [0]


def _msg_data_cmp(x, y):
    xl = x.get('data', _LIST_0)
    yl = y.get('data', _LIST_1)
    if xl == yl:
        return 0
    return -1 if xl < yl else 1


def _msg_cmp(x, y):
    xr = x.get('refersTo')
    yr = y.get('refersTo')
    if xr is None:
        if yr is None:
            return _msg_data_cmp(x, y)
        else:
            return 1
    if yr is None:
        return -1
    xri = xr.get('@idref')
    yri = yr.get('@idref')
    # _LOG.debug('xri = "{x}" yri = "{y}"'.format(x=xri, y=yri))
    if xri is None:
        if yri is None:
            xrk = list(xr.keys())
            xrk.sort()
            yrk = list(yr.keys())
            yrk.sort()
            if xrk == yrk:
                xrv = [xr[i] for i in xrk]
                yrv = [yr[i] for i in xrk]
                if xrv == yrv:
                    return _msg_data_cmp(x, y)
                if xrv < yrv:
                    return -1
                return 1
            return -1 if xrk < yrk else 1
        return 1
    if yri is None:
        return -1
    if xri == yri:
        return _msg_data_cmp(x, y)
    return -1 if xri < yri else 1


_msg_key_func = cmp_to_key(_msg_cmp)


class DefaultRichLogger(object):
    """A container for warnings and errors encountered during NexSON validation.

    `store_messages=False` argument to __init__ is *deprecated*. That argument
    is retained for backwards compatibility, but ignored. Messages are store
    regardless of the value of this argument.
    """
    def __init__(self, store_messages=False):
        self.out = sys.stderr
        self.store_messages_as_obj = store_messages
        self._warn_by_type = {}
        self._err_by_type = {}
        self._warn_by_obj = {}
        self._err_by_obj = {}
        self.prefix = ''
        self.retain_deprecated = False
        self.codes_to_skip = set()

    def has_error(self):
        """Returns `True` if no errors were stored."""
        return bool(self._err_by_type)

    @property
    def errors(self):
        """Returns the values of an internal errors by value dict.

        Returns an interable. Each element will be a
        set of tuples. The first element in each tuple will be
        a `MessageTupleAdaptor` instance.
        """
        return self._err_by_type.values()

    @property
    def warnings(self):
        """Returns the values of an internal warnings by value dict.

        Returns an interable. Each element will be a
        set of tuples. The first element in each tuple will be
        a `MessageTupleAdaptor` instance.
        """
        return self._warn_by_type.values()

    def is_logging_type(self, t):
        """Base class always returns `True`, as all messages are logged."""
        # pylint: disable=W0613,R0201
        return True

    def register_new_messages(self, err_tup, severity):
        """Stores a typed error tuple by type and python ID of target.

        The `err_tup` should be a tuple of the type described in `MessageTupleAdaptor`.
        The `severity` should be either `SeverityCodes.WARNING` or `SeverityCodes.ERROR`."""
        c = err_tup[0].code
        pyid = err_tup[1]
        if severity == SeverityCodes.WARNING:
            x = self._warn_by_type.setdefault(c, set())
            x.add(err_tup)
            x = self._warn_by_obj.setdefault(pyid, set())
            x.add(err_tup)
        else:
            x = self._err_by_type.setdefault(c, set())
            x.add(err_tup)
            x = self._err_by_obj.setdefault(pyid, set())
            x.add(err_tup)

    def get_err_warn_summary_dict(self, sort=True):
        """Returns a `{'warnings': {}, 'errors': {}}` view of all logged messages.

        The values in the warnings and errors dict will have a facet of
        `NexsonWarningCodes` as keys mapped to a list of the a slimmed version of the dict
        returned by calling `MessageTupleAdaptor.as_dict` on each logged warning or error.
        A key will be absent if no warning or error of that type was encountered.
        The "slimmed version" of the dict simply omits the '@code' value, because that will
        be the same for every error of each type (and the same as the key in the containing dict).
        """
        w = {}
        for wm in self._warn_by_type.values():
            d = _err_warn_summary(wm)
            w.update(d)
        e = {}
        for em in self._err_by_type.values():
            d = _err_warn_summary(em)
            e.update(d)
        if sort:
            for v in w.values():
                if isinstance(v, list):
                    v.sort(key=_msg_key_func)
            for v in e.values():
                if isinstance(v, list):
                    v.sort(key=_msg_key_func)
        return {'warnings': w, 'errors': e}

    def create_nexson_message_list(self, sort=True):
        """Returns a list of dict-forms of the log tuples described in `MessageTupleAdaptor`.

        The list elements are organized by logged event type.
        """
        em_list = []
        for key, em in self._err_by_type.items():
            d = _create_message_list(key, em, 'ERROR')
            em_list.extend(d)
        if sort:
            em_list.sort(key=_msg_key_func)
        wm_list = []
        for key, em in self._warn_by_type.items():
            d = _create_message_list(key, em, 'WARNING')
            wm_list.extend(d)
        if sort:
            em_list.sort(key=_msg_key_func)
            wm_list.sort(key=_msg_key_func)
        em_list.extend(wm_list)
        return em_list

    def prepare_annotation(self,
                           author_name='',
                           invocation=tuple(),
                           author_version=VERSION,
                           url='https://github.com/OpenTreeOfLife/peyotl',
                           description=None,
                           annotation_label=None  # @TEMP arg for backward compat.
                           ):
        """Wraps the set of logged warnings/errors as a dict, addable to a NexSON object.

        For example:

        .. highlight:: python
        .. code-block:: python

            obj = json.load(inp)
            v_log, adaptor = validate_nexson(obj)
            annotation = v_log.prepare_annotation(author_name=SCRIPT_NAME,
                                                  invocation=sys.argv[1:],
                                                  )
            adaptor.add_or_replace_annotation(obj,
                                              annotation['annotationEvent'],
                                              annotation['agent'],
                                              add_agent_only=args.add_agent_only)
        """
        if description is None:
            description = "validator of NexSON constraints as well as constraints " \
                          "that would allow a study to be imported into the Open Tree " \
                          "of Life's phylogenetic synthesis tools"
        # @TEMP. the args are in flux between the branches of api.opentreeoflife.org.
        #    which is bad. Hopefully we don't need annotation_label and
        #   can get rid of it.
        if annotation_label is not None:
            description += annotation_label
        # checks_performed = list(NexsonWarningCodes.numeric_codes_registered)
        # for code in self.codes_to_skip:
        #     try:
        #         checks_performed.remove(code)
        #     except:
        #         pass
        # checks_performed = [NexsonWarningCodes.facets[i] for i in checks_performed]
        agent_id = 'peyotl-validator'
        aevent_id = agent_id + '-event'
        ml = self.create_nexson_message_list()
        annotation = {
            '@id': aevent_id,
            '@description': description,
            '@wasAssociatedWithAgentId': agent_id,
            # '@dateCreated': datetime.datetime.utcnow().isoformat(),
            '@passedChecks': not self.has_error(),
            '@preserve': False,
            'message': ml
        }
        invocation_obj = {
            'commandLine': [i for i in invocation if i.startswith('--')],
            # 'checksPerformed': checks_performed,
            'otherProperty': [
                {'name': 'pythonVersion',
                 'value': platform.python_version()},
                {'name': 'pythonImplementation',
                 'value': platform.python_implementation(),
                 },
            ]
        }
        agent = {
            '@id': agent_id,
            '@name': author_name,
            '@url': url,
            '@description': description,
            '@version': author_version,
            'invocation': invocation_obj,
        }
        return {'annotationEvent': annotation, 'agent': agent}


class ValidationLogger(DefaultRichLogger):
    """Subclass of `DefaultRichLogger`. Retained for backward compatibility.

    No methods are overridden, simply uses base class behavior.
    """

    def __init__(self, store_messages=False):
        DefaultRichLogger.__init__(self, store_messages=store_messages)


class FilteringLogger(ValidationLogger):
    """Subclass of `DefaultRichLogger`.

    __init__ method allows client to supply `codes_to_register` or
    `codes_to_skip` ei
    """

    def __init__(self, codes_to_register=None, codes_to_skip=None, store_messages=False):
        ValidationLogger.__init__(self, store_messages=store_messages)
        self.codes_to_skip = set()
        if codes_to_register:
            self.registered = set(codes_to_register)
            if codes_to_skip:
                for el in codes_to_skip:
                    self.codes_to_skip.add(el)
                    assert el not in self.registered
        else:
            assert codes_to_skip
            self.registered = set(NexsonWarningCodes.numeric_codes_registered)
            for el in codes_to_skip:
                self.codes_to_skip.add(el)
                self.registered.remove(el)

    def is_logging_type(self, t):
        return (t not in self.codes_to_skip) and (t in self.registered)



#!/usr/bin/env python
"""
Library for handling the NexSON format for phylogenetic trees used by the phylesystem
portion of the Open Tree of Life project
"""
# Some imports to help our py2 code behave like py3
from __future__ import absolute_import, print_function, division

__version__ = '0.0.3'  # sync with setup.py
__all__ = ['syntax',
           'validation',
           'proxy']

from .syntax import (detect_nexson_version,
                     extract_otu_nexson,
                     extract_tree_nexson,
                     nexson_frag_write_newick,
                     )
from .validation import (ot_validate,
                         validate_nexson,
                         )
from .proxy import NexsonProxy

__all__ = [  # From syntax
           'detect_nexson_version',
           'extract_otu_nexson',
           'extract_tree_nexson',
           ]






#! /usr/bin/env python
from nexson.syntax import detect_nexson_version, get_empty_nexson
from peyutil import UNICODE
from nexson import validate_nexson
from nexson.test.support.pathmap import get_test_path_mapper
from peyutil.test.support.helper import testing_write_json as twj
from peyutil.test.support.helper import testing_read_json as trj
from peyutil.test.support.helper import testing_through_json as ttj
from peyutil.test.support.helper import testing_dict_eq as tde
from peyutil.test.support.helper import testing_conv_key_unicode_literal as tckul
import unittest
import logging
import os

pathmap = get_test_path_mapper()
_LOG = logging.getLogger(__name__)

# round trip filename tuples
VALID_NEXSON_DIRS = ['9', 'otu', ]


class TestConvert(unittest.TestCase):
    def testDetectVersion(self):
        o = pathmap.nexson_obj('invalid/bad_version.json.input')
        v = detect_nexson_version(o)
        self.assertEqual(v, '1.3.1')

    def testValidFilesPass(self):
        format_list = ['1.2']
        msg = ''
        for d in VALID_NEXSON_DIRS:
            for nf in format_list:
                frag = os.path.join(d, 'v{f}.json'.format(f=nf))
                nexson = pathmap.nexson_obj(frag)
                aa = validate_nexson(nexson)
                annot = aa[0]
                for e in annot.errors:
                    _LOG.debug('unexpected error from {f}: {m}'.format(f=frag, m=UNICODE(e)))
                if len(annot.errors) > 0:
                    ofn = pathmap.nexson_source_path(frag + '.output')
                    ew_dict = annot.get_err_warn_summary_dict()
                    twj(ew_dict, ofn)
                    msg = "File failed to validate cleanly. See {o}".format(o=ofn)
                self.assertEqual(len(annot.errors), 0, msg)

    def testInvalidFilesFail(self):
        msg = ''
        for fn in pathmap.all_files(os.path.join('nexson', 'invalid')):
            if fn.endswith('.input'):
                frag = fn[:-len('.input')]
                inp = trj(fn)
                try:
                    aa = validate_nexson(inp)
                except:
                    continue
                annot = aa[0]
                if len(annot.errors) == 0:
                    ofn = pathmap.nexson_source_path(frag + '.output')
                    ew_dict = annot.get_err_warn_summary_dict()
                    twj(ew_dict, ofn)
                    msg = "Failed to reject file. See {o}".format(o=str(msg))
                    self.assertTrue(False, msg)

    def testExpectedWarnings(self):
        msg = ''
        for fn in pathmap.all_files(os.path.join('nexson', 'warn_err')):
            if fn.endswith('.input'):
                frag = fn[:-len('.input')]
                efn = frag + '.expected'
                if os.path.exists(efn):
                    inp = trj(fn)
                    aa = validate_nexson(inp)
                    annot = aa[0]
                    ew_dict = annot.get_err_warn_summary_dict()
                    ew_dict = tckul(ttj(ew_dict))
                    exp = tckul(trj(efn))
                    if not tde(ew_dict, exp):
                        ofn = frag + '.output'
                        twj(ew_dict, ofn)
                        msg = "Validation failed to produce expected outcome. Compare {o} and {e}".format(o=ofn, e=efn)
                    self.assertDictEqual(exp, ew_dict, msg)
                else:
                    _LOG.warning('Expected output file "{f}" not found'.format(f=efn))

    def testOldExpectedWarnings(self):
        msg = ''
        for fn in pathmap.all_files(os.path.join('nexson', 'old-tests')):
            if fn.endswith('.input'):
                frag = fn[:-len('.input')]
                efn = frag + '.expected'
                if os.path.exists(efn):
                    inp = trj(fn)
                    aa = validate_nexson(inp)
                    annot = aa[0]
                    ew_dict = annot.get_err_warn_summary_dict()
                    ew_dict = ttj(ew_dict)
                    exp = trj(efn)
                    if not tde(ew_dict, exp):
                        ofn = frag + '.output'
                        twj(ew_dict, ofn)
                        msg = "Validation failed to produce expected outcome. Compare {o} and {e}".format(o=ofn, e=efn)
                    self.assertDictEqual(exp, ew_dict, msg)
                else:
                    _LOG.warning('Expected output file "{f}" not found'.format(f=efn))

    def testCreated(self):
        b = get_empty_nexson()
        aa = validate_nexson(b)
        annot = aa[0]
        self.assertFalse(annot.has_error())
        b = get_empty_nexson(include_cc0=True)
        aa = validate_nexson(b)
        annot = aa[0]
        self.assertFalse(annot.has_error())


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python
from nexson import validate_nexson
from peyutil.test.support.helper import testing_write_json as twj
from peyutil.test.support.helper import testing_read_json as trj
from nexson.test.support.pathmap import get_test_path_mapper
import unittest
import os

pathmap = get_test_path_mapper()

# round trip filename tuples
VALID_NEXSON_DIRS = ['9', 'otu', ]


class TestConvert(unittest.TestCase):
    def testInvalidFilesFail(self):
        msg = ''
        for fn in pathmap.all_files(os.path.join('nexson', 'lacking_otus')):
            if fn.endswith('.input'):
                frag = fn[:-len('.input')]
                inp = trj(fn)
                aa = validate_nexson(inp)
                annot = aa[0]
                # import json
                # print(json.dumps(annot.prepare_annotation(), indent=2))
                #for n, e in enumerate(annot.errors):
                    # print(n, type(e))
                    # for nv, v in enumerate(e):
                    #     vt = v[0]
                    #     print(nv, vt.as_dict(v))
                if len(annot.errors) == 0:
                    ofn = pathmap.nexson_source_path(frag + '.output')
                    ew_dict = annot.get_err_warn_summary_dict()
                    twj(ew_dict, ofn)
                    msg = "Failed to reject file. See {o}".format(o=str(msg))
                    self.assertTrue(False, msg)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""Simple manipulations of data structure in peyotl
"""
from __future__ import absolute_import, print_function, division
import logging
from .syntax import (_add_uniq_value_to_dict_bf,
                     BY_ID_HONEY_BADGERFISH,
                     convert_nexson_format,
                     count_num_trees,
                     detect_nexson_version,
                     get_nexml_el,
                     _is_by_id_hbf)

_LOG = logging.getLogger(__name__)


def iter_otus(nexson, nexson_version=None):
    """generator over all otus in all otus group elements.
    yields a tuple of 3 items:
        otus group ID,
        otu ID,
        the otu obj
    """
    if nexson_version is None:
        nexson_version = detect_nexson_version(nexson)
    if not _is_by_id_hbf(nexson_version):
        convert_nexson_format(nexson, BY_ID_HONEY_BADGERFISH)  # TODO shouldn't modify...
    nex = get_nexml_el(nexson)
    otus_group_by_id = nex['otusById']
    group_order = nex.get('^ot:otusElementOrder', [])
    if len(group_order) < len(otus_group_by_id):
        group_order = list(otus_group_by_id.keys())
        group_order.sort()
    for otus_group_id in group_order:
        otus_group = otus_group_by_id[otus_group_id]
        otu_by_id = otus_group['otuById']
        ti_order = list(otu_by_id.keys())
        for otu_id in ti_order:
            otu = otu_by_id[otu_id]
            yield otus_group_id, otu_id, otu


def iter_trees(nexson, nexson_version=None):
    """generator over all trees in all trees elements.
    yields a tuple of 3 items:
        trees element ID,
        tree ID,
        the tree obj
    """
    if nexson_version is None:
        nexson_version = detect_nexson_version(nexson)
    nex = get_nexml_el(nexson)
    if _is_by_id_hbf(nexson_version):
        trees_group_by_id = nex['treesById']
        group_order = nex.get('^ot:treesElementOrder', [])
        if len(group_order) < len(trees_group_by_id):
            group_order = list(trees_group_by_id.keys())
            group_order.sort()
        for trees_group_id in group_order:
            trees_group = trees_group_by_id[trees_group_id]
            tree_by_id = trees_group['treeById']
            ti_order = trees_group.get('^ot:treeElementOrder', [])
            if len(ti_order) < len(tree_by_id):
                ti_order = list(tree_by_id.keys())
                ti_order.sort()
            for tree_id in ti_order:
                tree = tree_by_id[tree_id]
                yield trees_group_id, tree_id, tree
    else:
        for trees_group in nex.get('trees', []):
            trees_group_id = trees_group['@id']
            for tree in trees_group.get('tree', []):
                tree_id = tree['@id']
                yield trees_group_id, tree_id, tree


def label_to_original_label_otu_by_id(otu_by_id):
    """Takes a v1.2 otuById dict and, for every otu,
    checks if ot:originalLabel exists. If it does not,
    but @label does, then ot:originalLabel is set to
    @label and @label is deleted.
    """
    for val in otu_by_id.values():
        orig = val.get('^ot:originalLabel')
        if orig is None:
            label = val.get('@label')
            if label:
                del val['@label']
                val['^ot:originalLabel'] = label


def replace_entity_references_in_meta_and_annotations(d, id2id):
    if isinstance(d, list):
        for el in d:
            replace_entity_references_in_meta_and_annotations(el, id2id)
    elif isinstance(d, dict):
        about_key = d.get('@about')
        try:
            if about_key and about_key.startswith('#'):
                s = about_key[1:]
                r = id2id.get(s)
                if r is not None:
                    d['@about'] = '#' + r
        except:
            pass
        for v in d.values():
            replace_entity_references_in_meta_and_annotations(v, id2id)


_special_otu_keys = frozenset(('@label', '^ot:originalLabel', '^ot:ottId', '^ot:ottTaxonName'))


def _merge_otu_do_not_fix_references(src, dest):
    for k in _special_otu_keys:
        if k not in dest and k in src:
            dest[k] = src[k]
    for k, v in src.items():
        if k not in _special_otu_keys:
            _add_uniq_value_to_dict_bf(dest, k, v)


def merge_otus_and_trees(nexson_blob):
    """Takes a nexson object:
        1. merges trees elements 2 - # trees into the first trees element.,
        2. merges otus elements 2 - # otus into the first otus element.
        3. if there is no ot:originalLabel field for any otu,
            it sets that field based on @label and deletes @label
        4. merges an otu elements using the rule:
              A. treat (ottId, originalLabel) as a key
              B. If otu objects in subsequent trees match originalLabel and
                have a matching or absent ot:ottId, then they are merged into
                the same OTUs (however see C)
              C. No two leaves of a tree may share an otu (though otu should
                be shared across different trees). It is important that
                each leaf node be mapped to a distinct OTU. Otherwise there
                will be no way of separating them during OTU mapping. we
                do this indirectly by assuring to no two otu objects in the
                same otus object get merged with each other (or to a common
                object)

        5. correct object references to deleted entities.

    This function is used to patch up NexSONs created by multiple imports, hence the
    substitution of '@label' for 'ot:originalLabel'. Ids are arbitrary for imports from
    non-nexml tools, so matching is done based on names. This should mimic the behavior
    of the analysis tools that produced the trees (for most/all such tools unique names
    constitute unique OTUs).
    """
    id_to_replace_id = {}
    orig_version = detect_nexson_version(nexson_blob)
    convert_nexson_format(nexson_blob, BY_ID_HONEY_BADGERFISH)
    nexson = get_nexml_el(nexson_blob)
    otus_group_order = nexson.get('^ot:otusElementOrder', [])
    # (ott, orig) -> list of otu elements
    retained_mapped2otu = {}
    # orig -> list of otu elements
    retained_orig2otu = {}
    # For the first (entirely retained) group of otus:
    #   1. assure that originalLabel is filled in
    #   2. register the otu in retained_mapped2otu and retained_orig2otu
    # otu elements that have no label, originalLabel or ottId will not
    #   be registered, so they'll never be matched.
    retained_ogi = None
    if len(otus_group_order) > 0:
        otus_group_by_id = nexson['otusById']
        retained_ogi = otus_group_order[0]
        retained_og = otus_group_by_id[retained_ogi]
        retained_og_otu = retained_og.setdefault('otuById', {})
        label_to_original_label_otu_by_id(retained_og_otu)
        for oid, otu in retained_og_otu.items():
            ottid = otu.get('^ot:ottId')
            orig = otu.get('^ot:originalLabel')
            key = (ottid, orig)
            if key != (None, None):
                m = retained_mapped2otu.setdefault(key, [])
                t = (oid, otu)
                m.append(t)
                if orig is not None:
                    m = retained_orig2otu.setdefault(orig, [])
                    m.append(t)
        # For each of the other otus elements, we:
        #   1. assure that originalLabel is filled in
        #   2. decide (for each otu) whether it will
        #       be added to retained_og or merged with
        #       an otu already in retained_og. In the
        #       case of the latter, we add to the
        #       replaced_otu dict (old oid as key, new otu as value)
        for ogi in otus_group_order[1:]:
            # _LOG.debug('retained_mapped2otu = {r}'.format(r=retained_mapped2otu))
            og = otus_group_by_id[ogi]
            del otus_group_by_id[ogi]
            otu_by_id = og.get('otuById', {})
            label_to_original_label_otu_by_id(otu_by_id)
            used_matches = set()
            id_to_replace_id[ogi] = retained_ogi
            for oid, otu in otu_by_id.items():
                ottid = otu.get('^ot:ottId')
                orig = otu.get('^ot:originalLabel')
                key = (ottid, orig)
                if key == (None, None):
                    retained_og[oid] = otu
                else:
                    match_otu = None
                    mlist = retained_mapped2otu.get(key)
                    if mlist is not None:
                        for m in mlist:
                            if m[0] not in used_matches:
                                # _LOG.debug('Matching {k} to {m}'.format(k=repr(key), m=repr(m)))
                                match_otu = m
                                break
                                # else:
                                #    _LOG.debug('{k} already in {m}'.format(k=repr(m[0]), m=repr(used_matches)))
                    if match_otu is None:
                        # _LOG.debug('New el: {k} mlist = {m}'.format(k=repr(key), m=repr(mlist)))
                        mlist = retained_orig2otu.get(orig, [])
                        for m in mlist:
                            if m[0] not in used_matches:
                                match_otu = m
                                break
                    if match_otu is not None:
                        id_to_replace_id[oid] = match_otu[0]
                        used_matches.add(match_otu[0])
                        _merge_otu_do_not_fix_references(otu, match_otu[1])
                    else:
                        assert oid not in retained_og_otu
                        retained_og_otu[oid] = otu
                        m = retained_mapped2otu.setdefault(key, [])
                        t = (oid, otu)
                        m.append(t)
                        if orig is not None:
                            m = retained_orig2otu.setdefault(orig, [])
                            m.append(t)
        nexson['^ot:otusElementOrder'] = [retained_ogi]
    # Move all of the tree elements to the first trees group.
    trees_group_order = nexson.get('^ot:treesElementOrder', [])
    if len(trees_group_order) > 0:
        assert retained_ogi is not None  # should not be able to get here with trees, but no OTUs
        trees_group_by_id = nexson['treesById']
        retained_tgi = trees_group_order[0]
        retained_tg = trees_group_by_id[retained_tgi]
        retained_tg['@otus'] = retained_ogi
        retained_tg_tree_obj = retained_tg.get('treeById', {})
        for tgi in trees_group_order[1:]:
            tg = trees_group_by_id[tgi]
            del trees_group_by_id[tgi]
            id_to_replace_id[tgi] = retained_tgi
            retained_tg['^ot:treeElementOrder'].extend(tg['^ot:treeElementOrder'])
            for tid, tree_obj in tg.get('treeById', {}).items():
                retained_tg_tree_obj[tid] = tree_obj
        for tree_obj in retained_tg_tree_obj.values():
            for node in tree_obj.get('nodeById', {}).values():
                o = node.get('@otu')
                if o is not None:
                    r = id_to_replace_id.get(o)
                    if r is not None:
                        node['@otu'] = r
        nexson['^ot:treesElementOrder'] = [retained_tgi]

    replace_entity_references_in_meta_and_annotations(nexson, id_to_replace_id)
    convert_nexson_format(nexson_blob, orig_version)
    return nexson_blob



#!/usr/bin/env
import logging

_LOG = logging.getLogger(__name__)


def error_return(msg):
    _LOG.debug(msg)
    return False


VERSION = '0.0.4a'


class SeverityCodes(object):
    """An enum of Warning/Error severity."""
    ERROR, WARNING = range(2)
    facets = ['ERROR', 'WARNING']
    numeric_codes_registered = set(range(len(facets)))


class _NEXEL(object):
    TOP_LEVEL = 0
    NEXML = 1
    OTUS = 2
    OTU = 3
    TREES = 4
    TREE = 5
    NODE = 6
    EDGE = 7
    META = 8
    INTERNAL_NODE = 9
    LEAF_NODE = 10
    CODE_TO_STR = {}
    CODE_TO_PAR_CODE = {}
    CODE_TO_OTHER_ID_KEY = {}
    CODE_TO_TOP_ENTITY_NAME = {}


_NEXEL.CODE_TO_STR = {
    _NEXEL.TOP_LEVEL: 'top-level',
    _NEXEL.NEXML: 'nexml',
    _NEXEL.OTUS: 'otus',
    _NEXEL.OTU: 'otu',
    _NEXEL.TREES: 'trees',
    _NEXEL.TREE: 'tree',
    _NEXEL.NODE: 'node',
    _NEXEL.INTERNAL_NODE: 'node',
    _NEXEL.LEAF_NODE: 'node',
    _NEXEL.EDGE: 'edge',
    _NEXEL.META: 'meta',
    None: 'unknown',
}
_NEXEL.CODE_TO_PAR_CODE = {
    _NEXEL.TOP_LEVEL: None,
    _NEXEL.NEXML: _NEXEL.TOP_LEVEL,
    _NEXEL.OTUS: _NEXEL.NEXML,
    _NEXEL.OTU: _NEXEL.OTUS,
    _NEXEL.TREES: _NEXEL.NEXML,
    _NEXEL.TREE: _NEXEL.TREES,
    _NEXEL.NODE: _NEXEL.TREE,
    _NEXEL.INTERNAL_NODE: _NEXEL.TREE,
    _NEXEL.LEAF_NODE: _NEXEL.TREE,
    _NEXEL.EDGE: _NEXEL.TREE,
    None: None,
}
_NEXEL.CODE_TO_OTHER_ID_KEY = {
    _NEXEL.TOP_LEVEL: None,
    _NEXEL.NEXML: None,
    _NEXEL.OTUS: '@otusID',
    _NEXEL.OTU: '@otuID',
    _NEXEL.TREES: '@treesID',
    _NEXEL.TREE: '@treeID',
    _NEXEL.NODE: '@nodeID',
    _NEXEL.INTERNAL_NODE: '@nodeID',
    _NEXEL.LEAF_NODE: '@nodeID',
    _NEXEL.EDGE: '@edgeID',
    None: None,
}
_NEXEL.CODE_TO_TOP_ENTITY_NAME = {
    _NEXEL.TOP_LEVEL: '',
    _NEXEL.NEXML: 'nexml',
    _NEXEL.OTUS: 'otus',
    _NEXEL.OTU: 'otus',
    _NEXEL.TREES: 'trees',
    _NEXEL.TREE: 'trees',
    _NEXEL.NODE: 'trees',
    _NEXEL.INTERNAL_NODE: 'trees',
    _NEXEL.LEAF_NODE: 'trees',
    _NEXEL.EDGE: 'trees',
    None: '',
}



#! /usr/bin/env python
import unittest
import logging
from nexson.manip import count_num_trees, iter_trees
from nexson.test.support.pathmap import get_test_path_mapper as ntest_path_mapper

pathmap = ntest_path_mapper()

_LOG = logging.getLogger(__name__)

class TestManip(unittest.TestCase):
    def testCanCountTrees(self):
        for v in ['0.0', '1.0', '1.2']:
            inp = pathmap.nexson_obj('otu/v{v}.json'.format(v=v))
            self.assertEqual(1, count_num_trees(inp, v))

    def testIterTree(self):
        id_order_list = []
        f_list = ['expected', 'input']
        for v in ['1.0', '1.2']:
            for f in f_list:
                inp = pathmap.nexson_obj('merge/merge-{f}.v{v}.json'.format(v=v, f=f))
                id_order = []
                for t_tuple in iter_trees(inp):
                    ti = t_tuple[1]
                    id_order.append(ti)
                id_order_list.append(id_order)
        for i in range(1, 4):
            self.assertEqual(id_order_list[0], id_order_list[i])


if __name__ == "__main__":
    unittest.main()



# !/usr/bin/env python
"""Functions for converting between the different representations
of NexSON and the NeXML representation.
See https://github.com/OpenTreeOfLife/api.opentreeoflife.org/wiki/NexSON

Most notable functions are:
    write_obj_as_nexml,
    get_ot_study_info_from_nexml,

"""
from nexson.syntax.helper import (ConversionConfig,
                                  NexsonConverter,
                                  NexsonError,
                                  detect_nexson_version,
                                  get_nexml_el,
                                  find_val_literal_meta_first,
                                  _add_uniq_value_to_dict_bf,
                                  _add_value_to_dict_bf,
                                  _get_index_list_of_values,
                                  _index_list_of_values,
                                  _is_badgerfish_version,
                                  _is_by_id_hbf,
                                  _is_direct_hbf,
                                  _is_supported_nexson_vers,
                                  BADGER_FISH_NEXSON_VERSION,
                                  DEFAULT_NEXSON_VERSION,
                                  DIRECT_HONEY_BADGERFISH,
                                  NEXML_NEXSON_VERSION,
                                  BY_ID_HONEY_BADGERFISH,
                                  SUPPORTED_NEXSON_VERSIONS,
                                  _simplify_all_meta_by_id_del,)
# noinspection PyUnresolvedReferences
from peyutil.input_output import read_as_json, write_as_json
from peyutil.str_util import (flush_utf_8_writer,
                              UNICODE,
                              is_str_type,
                              get_utf_8_string_io_writer)
from nexson.syntax.optimal2direct_nexson import Optimal2DirectNexson
from nexson.syntax.direct2optimal_nexson import Direct2OptimalNexson
from nexson.syntax.badgerfish2direct_nexson import Badgerfish2DirectNexson
from nexson.syntax.direct2badgerfish_nexson import Direct2BadgerfishNexson
from nexson.syntax.nexson2nexml import Nexson2Nexml
from nexson.syntax.nexml2nexson import Nexml2Nexson
from nexson.syntax.inspect import count_num_trees
import xml.dom.minidom
import codecs
import re
import logging

_CONVERTIBLE_FORMATS = frozenset([NEXML_NEXSON_VERSION,
                                  DIRECT_HONEY_BADGERFISH,
                                  BADGER_FISH_NEXSON_VERSION,
                                  BY_ID_HONEY_BADGERFISH,
                                  '0.0',
                                  '1.0',
                                  '1.2', ])
_LOG = logging.getLogger(__name__)


def iter_otu(nexson, nexson_version=None):
    if nexson_version is None:
        nexson_version = detect_nexson_version(nexson)
    if not _is_by_id_hbf(nexson_version):
        raise NotImplementedError('iter_otu is only supported for nexson 1.2 at this point')
    nexml = get_nexml_el(nexson)
    for og in nexml.get('otusById', {}).values():
        for otu_id, otu in og.get('otuById', {}).items():
            yield otu_id, otu


def strip_to_meta_only(blob, nexson_version):
    if nexson_version is None:
        nexson_version = detect_nexson_version(blob)
    nex = get_nexml_el(blob)
    if _is_by_id_hbf(nexson_version):
        for otus_group in nex.get('otusById', {}).values():
            if 'otuById' in otus_group:
                del otus_group['otuById']
        for trees_group in nex.get('treesById', {}).values():
            tree_group = trees_group['treeById']
            key_list = tree_group.keys()
            for k in key_list:
                tree_group[k] = None
    else:
        otus = nex['otus']
        if not isinstance(otus, list):
            otus = [otus]
        for otus_group in otus:
            if 'otu' in otus_group:
                del otus_group['otu']
        trees = nex['trees']
        if not isinstance(trees, list):
            trees = [trees]
        for trees_group in trees:
            tree_list = trees_group.get('tree')
            if not isinstance(tree_list, list):
                tree_list = [tree_list]
            t = [{'id': i.get('@id')} for i in tree_list]
            trees_group['tree'] = t


def _otu_dict_to_otumap(otu_dict):
    d = {}
    for v in otu_dict.values():
        k = v['^ot:originalLabel']
        mv = d.get(k)
        if mv is None:
            mv = {}
            d[k] = mv
        elif isinstance(mv, list):
            mv.append({})
            mv = mv[-1]
        else:
            mv = [mv, {}]
            mv = mv[-1]
            d[k] = mv
        for mk in ['^ot:ottId', '^ot:ottTaxonName']:
            mvv = v.get(mk)
            if mvv is not None:
                mv[mk] = mvv
    return d


def _get_content_id_from(content, **kwargs):
    if content in PhyloSchema._no_content_id_types:  # pylint: disable=W0212
        return None
    elif content == 'tree':
        return kwargs.get('tree_id')
    elif content == 'subtree':
        subtree_id = kwargs.get('subtree_id')
        if subtree_id is None:
            subtree_id = kwargs.get('node_id')
        return kwargs.get('tree_id'), subtree_id
    elif content == 'otus':
        return kwargs.get('otus_id')
    elif content == 'otu':
        return kwargs.get('otu_id')
    elif content == ('otus', 'otumap'):
        return kwargs.get('otus_id')


def _sniff_content_from_kwargs(**kwargs):
    c_id = kwargs.get('tree_id')
    if c_id is not None:
        s_id = kwargs.get('subtree_id')
        if s_id is None:
            s_id = kwargs.get('node_id')
        if s_id is None:
            return 'tree', c_id
        return 'subtree', (c_id, s_id)
    c_id = kwargs.get('otus_id')
    if c_id is not None:
        return 'otus', c_id
    c_id = kwargs.get('otu_id')
    if c_id is not None:
        return 'otu', c_id
    return 'study', None


def get_git_sha(blob):
    return blob['sha']


# noinspection PyUnusedLocal
def nodeid_labeller(nodeid, node, otu):
    return nodeid


# noinspection PyUnusedLocal
def nodeid_ottid_labeller(nodeid, node, otu):
    if otu is not None and '^ot:ottId' in otu.keys():
        ottid = otu['^ot:ottId']
        return "{}_ott{}".format(nodeid, ottid)
    else:
        return nodeid


# noinspection PyUnusedLocal
def _nodeid_ottid_labeller(nodeid, node, otu):
    if otu is not None and '^ot:ottId' in otu.keys():
        ottid = otu['^ot:ottId']
        return "_{}_ott{}".format(nodeid, ottid)
    else:
        return nodeid


def create_content_spec(**kwargs):
    """Sugar. factory for a PhyloSchema object.

    Repackages the kwargs to kwargs for PhyloSchema so that our
    PhyloSchema.__init__ does not have to be soo rich
    """
    format_str = kwargs.get('format', 'nexson')
    nexson_version = kwargs.get('nexson_version', 'native')
    otu_label = kwargs.get('otu_label')
    if otu_label is None:
        otu_label = kwargs.get('tip_label')
    content = kwargs.get('content')
    if content is not None:
        content_id = kwargs.get('content_id')
        if content_id is None:
            content_id = _get_content_id_from(**kwargs)
    else:
        content, content_id = _sniff_content_from_kwargs(**kwargs)
    if content is None:
        content = 'study'
    return PhyloSchema(content=content,
                       content_id=content_id,
                       format_str=format_str,
                       version=nexson_version,
                       otu_label=otu_label,
                       repo_nexml2json=kwargs.get('repo_nexml2json'),
                       bracket_ingroup=bool(kwargs.get('bracket_ingroup', False)),
                       cull_nonmatching=kwargs.get('cull_nonmatching'))


class PhyloSchema(object):
    """Simple container for holding the set of variables needed to
    convert from one format to another (with error checking).

    The primary motivation for this class is to:
        1. generate type conversion errors up front when some one requests
            a particular coercion. For example, this allows the phylesystem
            api to raise an error before it fetches the data in cases in which
            the user is requesting a format/content combination is not
            currently supported (or not possible)
        2. allow that agreed-upon coercion to be done later with a simple
            call to convert or serialize. So the class acts like a closure
            that can transform any nexson to the desired format (if NexSON
            has the necessary content)
    `bracket_ingroup` is currently only used in newick string export. If True, then
        [pre-ingroup-marker] and [post-ingroup-marker] comments will surround the ingroup
        definition in the tree.
    """
    _format_list = ('newick', 'nexson', 'nexml', 'nexus')
    NEWICK, NEXSON, NEXML, NEXUS = range(4)
    _extension2format = {
        '.nexson': 'nexson',
        '.nexml': 'nexml',
        '.nex': 'nexus',
        '.tre': 'newick',
        '.nwk': 'newick',
    }
    _otu_label2prop = {'ot:originallabel': '^ot:originalLabel',
                       'ot:ottid': '^ot:ottId',
                       'ot:otttaxonname': '^ot:ottTaxonName', }
    _otu_label_list = list(_otu_label2prop.keys()) + ['nodeid_ottid', '_nodeid_ottid', 'nodeid']
    _NEWICK_PROP_VALS = _otu_label2prop.values()
    _no_content_id_types = {'study', 'meta', 'treelist'}
    _tup_content_id_types = {'subtree'}
    _str_content_id_types = {'tree', 'otus', 'otu', 'otumap', 'file'}
    _content_types = {'file', 'study', 'tree', 'meta', 'otus', 'otu', 'otumap', 'subtree', 'treelist'}

    def __init__(self, schema=None, **kwargs):
        """Checks:
            'schema',
            'type_ext', then
            'output_nexml2json' (implicitly NexSON)
            If exporting to a non-nexson format, `otu_label` (and then
                `tip_label`are checked) to determine how to label the tips
                'ot:originallabel', 'ot:ottid', and 'ot:otttaxonname'
                are supported values
        """
        self.content = kwargs.get('content', 'study')
        self.bracket_ingroup = bool(kwargs.get('bracket_ingroup', False))
        self.content_id = kwargs.get('content_id')
        self.cull_nonmatching = kwargs.get('cull_nonmatching')
        err_msg = 'expected cull_nonmatching to be "true" or "false" or the boolean versions of those values. found {}'
        # pylint: disable=E1103
        if is_str_type(self.cull_nonmatching):
            if self.cull_nonmatching.lower() in ['true', '1']:
                self.cull_nonmatching = True
            else:
                if self.cull_nonmatching.lower() not in ['false', '0']:
                    raise ValueError(err_msg.format(kwargs.get('cull_nonmatching')))
                self.cull_nonmatching = False
        elif self.cull_nonmatching is not None and not isinstance(self.cull_nonmatching, bool):
            raise ValueError(err_msg.format('a non-boolean and non-string value'))

        if self.content not in PhyloSchema._content_types:
            raise ValueError('"content" must be one of: "{}"'.format('", "'.join(PhyloSchema._content_types)))
        if self.content in PhyloSchema._no_content_id_types:
            if self.content_id is not None:
                raise ValueError('No content_id expected for "{}" content'.format(self.content))
        elif self.content in PhyloSchema._str_content_id_types:
            if not (self.content_id is None or is_str_type(self.content_id)):
                raise ValueError('content_id for "{}" content must be a string (if provided)'.format(self.content))
        else:
            is_list = isinstance(self.content_id, list) or isinstance(self.content_id, tuple)
            if (self.content_id is None) or (not is_list) or len(self.content_id) != 2:
                raise ValueError('Expecting 2 content_ids for the "subtree" content')
        if schema is not None:
            # _LOG.debug('schema from schema arg')
            self.format_str = schema.lower()
        elif kwargs.get('type_ext') is not None:
            # _LOG.debug('schema from type_ext arg')
            ext = kwargs['type_ext'].lower()
            try:
                self.format_str = PhyloSchema._extension2format[ext]
            except:
                raise ValueError('file extension "{}" not recognized'.format(kwargs['type_ext']))
        elif kwargs.get('output_nexml2json') is not None:
            # _LOG.debug('schema from output_nexml2json arg')
            self.format_str = 'nexson'
            self.version = kwargs['output_nexml2json']
        else:
            # _LOG.debug('schema from format_str arg')
            self.format_str = kwargs.get('format_str')
        if self.format_str is None:
            raise ValueError('Expecting "format_str", "schema", or "type_ext" argument')
        try:
            # _LOG.debug('self.format_str = {}'.format(self.format_str))
            self.format_code = PhyloSchema._format_list.index(self.format_str)
            # _LOG.debug('self.format_code = {}'.format(str(self.format_code)))
        except:
            raise ValueError('format "{}" not recognized'.format(self.format_str))
        if self.format_code == PhyloSchema.NEXSON:
            try:
                if not hasattr(self, 'version'):
                    if 'output_nexml2json' in kwargs:
                        self.version = kwargs['output_nexml2json']
                    else:
                        self.version = kwargs['version']
                if self.version == 'native':
                    self.version = kwargs['repo_nexml2json']
                if not _is_supported_nexson_vers(self.version):
                    raise ValueError('The "{}" version of NexSON is not supported'.format(self.version))
            except:
                msg = 'Expecting version of NexSON to be specified using ' \
                      '"output_nexml2json" argument (or via some other mechanism)'
                raise ValueError(msg)
        else:
            if self.content in ['meta']:
                raise ValueError('The "{}" content can only be returned in NexSON'.format(self.content))
            if kwargs.get('otu_label') is not None:
                self.otu_label = kwargs['otu_label'].lower()
            else:
                self.otu_label = kwargs.get('tip_label', 'ot:originallabel').lower()
            if self.otu_label not in PhyloSchema._otu_label_list:
                with_ns = 'ot:{}'.format(self.otu_label)
                if with_ns in PhyloSchema._otu_label_list:
                    self.otu_label = with_ns
                else:
                    m = '"otu_label" or "tip_label" must be one of "{}"'
                    m = m.format('", "'.join(PhyloSchema._otu_label_list))
                    raise ValueError(m)
            if self.otu_label == "nodeid_ottid":
                self.otu_label_prop = nodeid_ottid_labeller
            elif self.otu_label == "_nodeid_ottid":
                self.otu_label_prop = _nodeid_ottid_labeller
            elif self.otu_label == "nodeid":
                self.otu_label_prop = nodeid_labeller
            else:
                self.otu_label_prop = PhyloSchema._otu_label2prop[self.otu_label]

    @property
    def description(self):
        if self.format_code == PhyloSchema.NEXSON:
            return 'NexSON v{v}'.format(v=self.version)
        elif self.format_code == PhyloSchema.NEXML:
            return 'NeXML'
        elif self.format_code == PhyloSchema.NEXUS:
            return 'NEXUS'
        elif self.format_code == PhyloSchema.NEWICK:
            return 'Newick'

    # noinspection PyUnusedLocal
    def can_convert_from(self, src_schema=None):  # pylint: disable=W0613
        if self.format_code == PhyloSchema.NEXSON:
            return self.content != 'subtree'
        if self.content == 'study':
            return True
        if self.content in {'tree', 'subtree'}:
            return self.format_code in [PhyloSchema.NEWICK, PhyloSchema.NEXUS]
        return False

    def is_json(self):
        return self.format_code == PhyloSchema.NEXSON

    def is_xml(self):
        return self.format_code == PhyloSchema.NEXML

    def is_text(self):
        return self.format_code in (PhyloSchema.NEXUS, PhyloSchema.NEWICK)

    def _phylesystem_api_params(self):
        d = {}
        if self.format_code == PhyloSchema.NEXSON:
            d['output_nexml2json'] = self.version
        else:
            if self.otu_label != 'ot:originallabel':
                d['otu_label'] = self.otu_label
        if self.content == 'tree':
            if self.cull_nonmatching:
                d['cull_nonmatching'] = 'true'
        return d

    def _phylesystem_api_ext(self):
        if self.format_code == PhyloSchema.NEXSON:
            return ''
        elif self.format_code == PhyloSchema.NEWICK:
            return '.tre'
        elif self.format_code == PhyloSchema.NEXUS:
            return '.nex'
        elif self.format_code == PhyloSchema.NEXML:
            return '.nexml'
        else:
            assert False

    def phylesystem_api_url(self, base_url, study_id):
        """Returns URL and param dict for a GET call to phylesystem_api
        """
        p = self._phylesystem_api_params()
        e = self._phylesystem_api_ext()
        if self.content == 'study':
            return '{d}/study/{i}{e}'.format(d=base_url, i=study_id, e=e), p
        elif self.content == 'tree':
            if self.content_id is None:
                return '{d}/study/{i}/tree{e}'.format(d=base_url, i=study_id, e=e), p
            return '{d}/study/{i}/tree/{t}{e}'.format(d=base_url, i=study_id, t=self.content_id, e=e), p
        elif self.content == 'subtree':
            assert self.content_id is not None
            t, n = self.content_id
            p['subtree_id'] = n
            return '{d}/study/{i}/subtree/{t}{e}'.format(d=base_url, i=study_id, t=t, e=e), p
        elif self.content == 'meta':
            return '{d}/study/{i}/meta{e}'.format(d=base_url, i=study_id, e=e), p
        elif self.content == 'otus':
            if self.content_id is None:
                return '{d}/study/{i}/otus{e}'.format(d=base_url, i=study_id, e=e), p
            return '{d}/study/{i}/otus/{t}{e}'.format(d=base_url, i=study_id, t=self.content_id, e=e), p
        elif self.content == 'otu':
            if self.content_id is None:
                return '{d}/study/{i}/otu{e}'.format(d=base_url, i=study_id, e=e), p
            return '{d}/study/{i}/otu/{t}{e}'.format(d=base_url, i=study_id, t=self.content_id, e=e), p
        elif self.content == 'otumap':
            return '{d}/otumap/{i}{e}'.format(d=base_url, i=study_id, e=e), p
        else:
            assert False

    def serialize(self, src, output_dest=None, src_schema=None):
        return self.convert(src, serialize=True, output_dest=output_dest, src_schema=src_schema)

    def convert(self, src, serialize=None, output_dest=None, src_schema=None):
        if src_schema is None:
            src_format = PhyloSchema.NEXSON
            current_format = None
        else:
            src_format = src_schema.format_code
            current_format = src_schema.version
        if not self.can_convert_from():
            m = 'Conversion of {c} to {d} is not supported'.format(c=self.content, d=self.description)
            raise NotImplementedError(m)
        if src_format != PhyloSchema.NEXSON:
            raise NotImplementedError('Only conversion from NexSON is currently supported')
        if self.format_code == PhyloSchema.NEXSON:
            d = src
            if self.content == 'study':
                d = convert_nexson_format(src,
                                          out_nexson_format=self.version,
                                          current_format=current_format,
                                          remove_old_structs=True,
                                          pristine_if_invalid=False,
                                          sort_arbitrary=False)
            elif self.content in ('tree', 'subtree'):
                if self.content == 'tree' and self.cull_nonmatching:
                    d = cull_nonmatching_trees(d, self.content_id, current_format)
                    d = convert_nexson_format(d,
                                              out_nexson_format=self.version,
                                              current_format=current_format,
                                              remove_old_structs=True,
                                              pristine_if_invalid=False,
                                              sort_arbitrary=False)

                else:
                    i_t_o_list = extract_tree_nexson(d, self.content_id, current_format)
                    d = {}
                    for ito_tup in i_t_o_list:
                        i, t = ito_tup[0], ito_tup[1]
                        d[i] = t
            elif self.content == 'meta':
                strip_to_meta_only(d, current_format)
            elif self.content == 'otus':
                d = extract_otus_nexson(d, self.content_id, current_format)
            elif self.content == 'otu':
                d = extract_otu_nexson(d, self.content_id, current_format)
            elif self.content == 'otumap':
                if self.content_id is None:
                    r = extract_otu_nexson(d, None, current_format)
                else:
                    p = extract_otus_nexson(d, self.content_id, current_format)
                    if p is None:
                        r = extract_otu_nexson(d, self.content_id, current_format)
                    else:
                        r = {}
                        for v in p.values():
                            r.update(v.get('otuById', {}))
                if not r:
                    return None
                d = _otu_dict_to_otumap(r)
            elif self.content == 'treelist':
                i_t_o_list = extract_tree_nexson(d,
                                                 self.content_id,
                                                 current_format)
                d = [i[0] for i in i_t_o_list]
            if d is None:
                return None
            if serialize:
                if output_dest:
                    write_as_json(d, output_dest)
                    return None
                else:
                    f, wrapper = get_utf_8_string_io_writer()
                    write_as_json(d, wrapper)
                    flush_utf_8_writer(wrapper)
                    return f.getvalue()
            else:
                return d
        # Non-NexSON types go here...
        if (serialize is not None) and (not serialize):
            raise ValueError('Conversion without serialization is only supported for the NexSON format')
        if output_dest:
            if is_str_type(output_dest):
                output_dest = codecs.open(output_dest, 'w', encoding='utf-8')
        if self.format_code == PhyloSchema.NEXML:
            if output_dest:
                write_obj_as_nexml(src, output_dest, addindent=' ', newl='\n', otu_label=self.otu_label_prop)
                return
            return convert_to_nexml(src, addindent=' ', newl='\n', otu_label=self.otu_label_prop)
        elif self.format_code in [PhyloSchema.NEXUS, PhyloSchema.NEWICK]:
            if self.content in ('tree', 'subtree'):
                if isinstance(self.content_id, list) or isinstance(self.content_id, tuple):
                    ci, subtree_id = self.content_id
                else:
                    ci, subtree_id = self.content_id, None
            else:
                ci, subtree_id = None, None
            response = extract_tree(src, ci, self, subtree_id=subtree_id)
            # these formats are always serialized...
            if output_dest:
                output_dest.write(response)
                output_dest.write('\n')
            return response
        assert False


def get_ot_study_info_from_nexml(src=None,
                                 nexml_content=None,
                                 encoding=u'utf8',
                                 nexson_syntax_version=DEFAULT_NEXSON_VERSION):
    """Converts an XML doc to JSON using the honeybadgerfish convention (see to_honeybadgerfish_dict)
    and then prunes elements not used by open tree of life study curartion.

    If nexml_content is provided, it is interpreted as the contents
    of an NeXML file in utf-8 encoding.

    If nexml_content is None, then the src arg will be used src can be either:
        * a file_object, or
        * a string
    If `src` is a string then it will be treated as a filepath unless it
        begins with http:// or https:// (in which case it will be downloaded
        using peyutil.download)
    Returns a dictionary with the keys/values encoded according to the honeybadgerfish convention
    See https://github.com/OpenTreeOfLife/api.opentreeoflife.org/wiki/HoneyBadgerFish

    Currently:
        removes nexml/characters @TODO: should replace it with a URI for
            where the removed character data can be found.
    """
    if _is_by_id_hbf(nexson_syntax_version):
        nsv = DIRECT_HONEY_BADGERFISH
    else:
        nsv = nexson_syntax_version
    if nexml_content is None:
        if is_str_type(src):
            if src.startswith('http://') or src.startswith('https://'):
                from peyutil import download
                nexml_content = download(url=src, encoding=encoding)
                nexml_content = nexml_content.encode('utf-8')
            else:
                with codecs.open(src, 'r', encoding=encoding) as src:
                    nexml_content = src.read().encode('utf-8')
        else:
            nexml_content = src.read().encode('utf-8')
    doc = xml.dom.minidom.parseString(nexml_content)
    doc_root = doc.documentElement

    ccfg = ConversionConfig(output_format=nsv, input_format=NEXML_NEXSON_VERSION)
    converter = Nexml2Nexson(ccfg)
    o = converter.convert(doc_root)
    if _is_by_id_hbf(nexson_syntax_version):
        o = convert_nexson_format(o, BY_ID_HONEY_BADGERFISH, current_format=nsv)
    if 'nex:nexml' in o:
        n = o['nex:nexml']
        del o['nex:nexml']
        o['nexml'] = n
    return o


def _nexson_directly_translatable_to_nexml(vers):
    """TEMP: until we refactor nexml writing code to be more general..."""
    return (_is_badgerfish_version(vers)
            or _is_direct_hbf(vers)
            or vers == 'nexml')


def write_obj_as_nexml(obj_dict,
                       file_obj,
                       addindent='',
                       newl='',
                       use_default_root_atts=True,
                       otu_label='ot:originalLabel'):
    nsv = detect_nexson_version(obj_dict)
    if not _nexson_directly_translatable_to_nexml(nsv):
        convert_nexson_format(obj_dict, DIRECT_HONEY_BADGERFISH)
        nsv = DIRECT_HONEY_BADGERFISH
    ccfg = ConversionConfig(NEXML_NEXSON_VERSION,
                            input_format=nsv,
                            use_default_root_atts=use_default_root_atts,
                            otu_label=otu_label)
    converter = Nexson2Nexml(ccfg)
    doc = converter.convert(obj_dict)
    doc.writexml(file_obj, addindent=addindent, newl=newl, encoding='utf-8')


def convert_to_nexml(obj_dict, addindent='', newl='', use_default_root_atts=True, otu_label='ot:originalLabel'):
    f, wrapper = get_utf_8_string_io_writer()
    write_obj_as_nexml(obj_dict,
                       file_obj=wrapper,
                       addindent=addindent,
                       newl=newl,
                       use_default_root_atts=use_default_root_atts,
                       otu_label=otu_label)
    flush_utf_8_writer(wrapper)
    return f.getvalue()


def resolve_nexson_format(v):
    if len(v) == 3:
        if v == '1.2':
            return BY_ID_HONEY_BADGERFISH
        if v == '1.0':
            return DIRECT_HONEY_BADGERFISH
        if v == '0.0':
            return BADGER_FISH_NEXSON_VERSION
    else:
        if v in SUPPORTED_NEXSON_VERSIONS:
            return v
    raise NotImplementedError('NexSON version "{v}" not supported.'.format(v=v))


def can_convert_nexson_forms(src_format, dest_format):
    return (dest_format in _CONVERTIBLE_FORMATS) and (src_format in _CONVERTIBLE_FORMATS)


def convert_nexson_format(blob,
                          out_nexson_format,
                          current_format=None,
                          remove_old_structs=True,
                          pristine_if_invalid=False,
                          sort_arbitrary=False):
    """Take a dict form of NexSON and converts its datastructures to
    those needed to serialize as out_nexson_format.
    If current_format is not specified, it will be inferred.
    If `remove_old_structs` is False and different honeybadgerfish varieties
        are selected, the `blob` will be 'fat" containing both types
        of lookup structures.
    If pristine_if_invalid is False, then the object may be corrupted if it
        is an invalid nexson struct. Setting this to False can result in
        faster translation, but if an exception is raised the object may
        be polluted with partially constructed fields for the out_nexson_format.
    """
    if not current_format:
        current_format = detect_nexson_version(blob)
    out_nexson_format = resolve_nexson_format(out_nexson_format)
    if current_format == out_nexson_format:
        if sort_arbitrary:
            sort_arbitrarily_ordered_nexson(blob)
        return blob
    two2zero = _is_by_id_hbf(out_nexson_format) and _is_badgerfish_version(current_format)
    zero2two = _is_by_id_hbf(current_format) and _is_badgerfish_version(out_nexson_format)
    if two2zero or zero2two:
        # go from 0.0 -> 1.0 then the 1.0->1.2 should succeed without nexml...
        blob = convert_nexson_format(blob,
                                     DIRECT_HONEY_BADGERFISH,
                                     current_format=current_format,
                                     remove_old_structs=remove_old_structs,
                                     pristine_if_invalid=pristine_if_invalid)
        current_format = DIRECT_HONEY_BADGERFISH
    ccdict = {'output_format': out_nexson_format,
              'input_format': current_format,
              'remove_old_structs': remove_old_structs,
              'pristine_if_invalid': pristine_if_invalid}
    ccfg = ConversionConfig(ccdict)
    if _is_badgerfish_version(current_format):
        converter = Badgerfish2DirectNexson(ccfg)
    elif _is_badgerfish_version(out_nexson_format):
        assert _is_direct_hbf(current_format)
        converter = Direct2BadgerfishNexson(ccfg)
    elif _is_direct_hbf(current_format) and (out_nexson_format == BY_ID_HONEY_BADGERFISH):
        converter = Direct2OptimalNexson(ccfg)
    elif _is_direct_hbf(out_nexson_format) and (current_format == BY_ID_HONEY_BADGERFISH):
        converter = Optimal2DirectNexson(ccfg)
    else:
        raise NotImplementedError('Conversion from {i} to {o}'.format(i=current_format, o=out_nexson_format))
    blob = converter.convert(blob)
    if sort_arbitrary:
        sort_arbitrarily_ordered_nexson(blob)
    return blob


def _recursive_sort_meta(blob, k):
    # _LOG.debug('k=' + k)
    if isinstance(blob, list):
        for i in blob:
            if isinstance(i, list) or isinstance(i, dict):
                _recursive_sort_meta(i, k)
    else:
        for inner_k, v in blob.items():
            if inner_k == 'meta' and isinstance(v, list):
                sl = []
                incd = {}
                for el in v:
                    sk = el.get('@property') or el.get('@rel') or ''
                    count = incd.setdefault(sk, 0)
                    incd[sk] = 1 + count
                    sl.append((sk, count, el))
                sl.sort()
                del v[:]  # clear out the value in place
                v.extend([i[2] for i in sl])  # replace it with the item from the sorted list
            if isinstance(v, list) or isinstance(v, dict):
                _recursive_sort_meta(v, inner_k)


def sort_meta_elements(blob):
    """For v0.0 (which has meta values in a list), this
    function recursively walks through the object
    and sorts each meta by @property or @rel values.
    """
    v = detect_nexson_version(blob)
    if _is_badgerfish_version(v):
        _recursive_sort_meta(blob, '')
    return blob


def _inplace_sort_by_id(unsorted_list):
    """Takes a list of dicts each of which has an '@id' key,
    sorts the elements in the list by the value of the @id key.
    Assumes that @id is unique or the dicts have a meaningul < operator
    """
    if not isinstance(unsorted_list, list):
        return
    sorted_list = [(i.get('@id'), i) for i in unsorted_list]
    sorted_list.sort()
    del unsorted_list[:]
    unsorted_list.extend([i[1] for i in sorted_list])


def sort_arbitrarily_ordered_nexson(blob):
    """Primarily used for testing (getting nice diffs). Calls
    sort_meta_elements and then sorts otu, node and edge list by id
    """
    # otu, node and edge elements have no necessary orger in v0.0 or v1.0
    v = detect_nexson_version(blob)
    nex = get_nexml_el(blob)
    if _is_by_id_hbf(v):
        return blob
    sort_meta_elements(blob)
    for ob in _get_index_list_of_values(nex, 'otus'):
        _inplace_sort_by_id(ob.get('otu', []))
    for tb in _get_index_list_of_values(nex, 'trees'):
        for tree in _get_index_list_of_values(tb, 'tree'):
            _inplace_sort_by_id(tree.get('node', []))
            _inplace_sort_by_id(tree.get('edge', []))
    return blob


def add_resource_meta(obj, rel, href, version):
    if _is_badgerfish_version(version):
        m = obj.setdefault('meta', [])
        if not isinstance(m, list):
            m = [m]
            obj['meta'] = m
        m.append({'@href': href,
                  '@rel': rel,
                  '@xsi:type': 'nex:ResourceMeta'})
    else:
        k = '^' + rel
        _add_value_to_dict_bf(obj, k, {'@href': href})


def get_empty_nexson(vers='1.2.1', include_cc0=False):
    assert vers == '1.2.1'
    nexson = {
        'nexml': {
            '@about': '#study',
            '@generator': 'Open Tree API',
            '@id': 'study',
            '@nexml2json': vers,
            '@nexmljson': 'http://purl.org/opentree/nexson',
            '@version': '0.9',
            '@xmlns': {
                '$': 'http://www.nexml.org/2009',
                'nex': 'http://www.nexml.org/2009',
                'ot': 'http://purl.org/opentree-terms#',
                'xsd': 'http://www.w3.org/2001/XMLSchema#',
                'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
                'xhtml': 'http://www.w3.org/1999/xhtml/vocab#'
            },
            '^ot:otusElementOrder': [
                'otus1',
            ],
            'otusById': {
                'otus1': {
                    'otuById': {},
                },
            },
            '^ot:treesElementOrder': [
                'trees1',
            ],
            'treesById': {
                'trees1': {
                    '@otus': 'otus1',
                    '^ot:treeElementOrder': [],
                    'treeById': {},
                },
            },
        }
    }
    # N.B. We no longer require the CC0 waiver, so we should not assume it's here
    if include_cc0:
        nexson['nexml']['^xhtml:license'] = {'@href': 'http://creativecommons.org/publicdomain/zero/1.0/'}
    return nexson


_EMPTY_TUPLE = tuple
NEWICK_NEEDING_QUOTING = re.compile(r'(\s|[\[\]():,;])')
NEXUS_NEEDING_QUOTING = re.compile(r'(\s|[-()\[\]{}/\\,;:=*"`+<>])')


def quote_newick_name(s, needs_quotes_pattern=NEWICK_NEEDING_QUOTING):
    s = UNICODE(s)
    if "'" in s:
        return u"'{}'".format("''".join(s.split("'")))
    if needs_quotes_pattern.search(s):
        return u"'{}'".format(s)
    return s


def _write_newick_leaf_label(out, node_id, node, otu_group, label_key, leaf_labels, unlabeled_counter,
                             needs_quotes_pattern):
    """
    `label_key` is a string (a key in the otu object) or a callable that takes two arguments: the node, and the otu
    If `leaf_labels` is not None, it shoulr be a (list, dict) pair which will be filled. The list will
        hold the order encountered,
        and the dict will map name to index in the list
    """
    otu_id = node['@otu']
    otu = otu_group[otu_id]
    if is_str_type(label_key):
        label = otu.get(label_key)
        if label is None:
            unlabeled_counter += 1
            o = otu.get('^ot:originalLabel', '<unknown>')
            label = "'*tip #{n:d} not mapped to OTT. Original label - {o}'"
            label = label.format(n=unlabeled_counter, o=o)
        else:
            label = quote_newick_name(label, needs_quotes_pattern)
    else:
        label = quote_newick_name(label_key(node_id, node, otu), needs_quotes_pattern)
    if leaf_labels is not None:
        if label not in leaf_labels[1]:
            leaf_labels[1][label] = len(leaf_labels[0])
            leaf_labels[0].append(label)
    out.write(label)
    return unlabeled_counter


def _write_newick_internal_label(out, node_id, node, otu_group, label_key, needs_quotes_pattern):
    """`label_key` is a string (a key in the otu object) or a callable that takes two arguments:
        the node, and the otu (which may be None for an internal node)
    If `leaf_labels` is not None, it shoulr be a (list, dict) pair which will be filled. The list will
        hold the order encountered,
        and the dict will map name to index in the list
    """
    otu_id = node.get('@otu')
    if is_str_type(label_key):
        if otu_id is None:
            return
        otu = otu_group[otu_id]
        label = otu.get(label_key)
    else:
        label = label_key(node_id, node, None)
    if label is not None:
        label = quote_newick_name(label, needs_quotes_pattern)
        out.write(label)


def _write_newick_edge_len(out, edge):
    if edge is None:
        return
    e_len = edge.get('@length')
    if e_len is not None:
        out.write(':{e}'.format(e=e_len))


def convert_tree_to_newick(tree,
                           otu_group,
                           label_key,
                           leaf_labels,
                           needs_quotes_pattern=NEWICK_NEEDING_QUOTING,
                           subtree_id=None,
                           bracket_ingroup=False):
    """`label_key` is a string (a key in the otu object) or a callable that takes two arguments:
        the node, and the otu (which may be None for an internal node)
    If `leaf_labels` is not None, it shoulr be a (list, dict) pair which will be filled. The list will
        hold the order encountered,
        and the dict will map name to index in the list
    """
    assert (not is_str_type(label_key)) or (label_key in PhyloSchema._NEWICK_PROP_VALS)  # pylint: disable=W0212
    ingroup_node_id = tree.get('^ot:inGroupClade')
    if subtree_id:
        if subtree_id == 'ingroup':
            root_id = ingroup_node_id
            ingroup_node_id = None  # turns of the comment pre-ingroup-marker
        else:
            root_id = subtree_id
    else:
        root_id = tree['^ot:rootNodeId']
    edges = tree['edgeBySourceId']
    if root_id not in edges:
        return None
    nodes = tree['nodeById']
    sio, out = get_utf_8_string_io_writer()
    nexson_frag_write_newick(out,
                             edges,
                             nodes,
                             otu_group,
                             label_key,
                             leaf_labels,
                             root_id,
                             needs_quotes_pattern=needs_quotes_pattern,
                             ingroup_id=ingroup_node_id,
                             bracket_ingroup=bracket_ingroup)
    flush_utf_8_writer(out)
    return sio.getvalue()


def get_subtree_otus(nexson, tree_id, subtree_id=None, return_format='otu_id'):
    assert (return_format in ['otu_id', 'ottid'])
    tree = extract_tree_nexson(nexson, tree_id)[0][1]
    ingroup_node_id = tree.get('^ot:inGroupClade')
    edges = tree['edgeBySourceId']
    nodes = tree['nodeById']
    if subtree_id:
        if subtree_id == 'ingroup':
            root_id = ingroup_node_id
        else:
            root_id = subtree_id
    else:
        root_id = tree['^ot:rootNodeId']
    if root_id not in edges:
        return None
    otuset = set()
    todo = set()
    todo.add(root_id)
    while todo:
        curr_node_id = todo.pop()
        outgoing_edges = edges.get(curr_node_id)
        if outgoing_edges is None:
            otu = nodes.get(curr_node_id).get('@otu')
            assert otu
            if return_format == 'otu_id':
                otuset.add(otu)
            if return_format == 'ottid':
                d = extract_otu_nexson(nexson, otu, detect_nexson_version(nexson))
                ottid = d[otu].get('^ot:ottId')
                otuset.add(ottid)
        else:
            for edge, info in outgoing_edges.items():
                todo.add(info.get('@target'))
    return otuset


def nexson_frag_write_newick(out,
                             edges,
                             nodes,
                             otu_group,
                             label_key,
                             leaf_labels,
                             root_id,
                             needs_quotes_pattern=NEWICK_NEEDING_QUOTING,
                             ingroup_id=None,
                             bracket_ingroup=False,
                             with_edge_lengths=True):
    """`label_key` is a string (a key in the otu object) or a callable that takes two arguments:
        the node, and the otu (which may be None for an internal node)
    If `leaf_labels` is not None, it shoulr be a (list, dict) pair which will be filled. The list will
        hold the order encountered,
        and the dict will map name to index in the list
    """
    unlabeled_counter = 0
    curr_node_id = root_id
    assert curr_node_id
    curr_edge = None
    curr_sib_list = []
    curr_stack = []
    going_tipward = True
    while True:
        if going_tipward:
            outgoing_edges = edges.get(curr_node_id)
            if outgoing_edges is None:
                curr_node = nodes[curr_node_id]
                assert curr_node_id is not None
                assert curr_node_id is not None
                unlabeled_counter = _write_newick_leaf_label(out,
                                                             curr_node_id,
                                                             curr_node,
                                                             otu_group,
                                                             label_key,
                                                             leaf_labels,
                                                             unlabeled_counter,
                                                             needs_quotes_pattern)
                if with_edge_lengths:
                    _write_newick_edge_len(out, curr_edge)
                going_tipward = False
            else:
                te = [(i, e) for i, e in outgoing_edges.items()]
                te.sort()  # produce a consistent rotation... Necessary?
                if bracket_ingroup and (ingroup_id == curr_node_id):
                    out.write('[pre-ingroup-marker]')
                out.write('(')
                next_p = te.pop(0)
                curr_stack.append((curr_edge, curr_node_id, curr_sib_list))
                curr_edge, curr_sib_list = next_p[1], te
                curr_node_id = curr_edge['@target']
        if not going_tipward:
            next_up_edge_id = None
            while True:
                if curr_sib_list:
                    out.write(',')
                    next_up_edge_id, next_up_edge = curr_sib_list.pop(0)
                    break
                if curr_stack:
                    curr_edge, curr_node_id, curr_sib_list = curr_stack.pop(-1)
                    curr_node = nodes[curr_node_id]
                    out.write(')')
                    _write_newick_internal_label(out,
                                                 curr_node_id,
                                                 curr_node,
                                                 otu_group,
                                                 label_key,
                                                 needs_quotes_pattern)
                    if with_edge_lengths:
                        _write_newick_edge_len(out, curr_edge)
                    if bracket_ingroup and (ingroup_id == curr_node_id):
                        out.write('[post-ingroup-marker]')
                else:
                    break
            if next_up_edge_id is None:
                break
            # noinspection PyUnboundLocalVariable
            curr_edge = next_up_edge
            curr_node_id = curr_edge['@target']
            going_tipward = True
    out.write(';')


def _write_nexus_format(quoted_leaf_labels, tree_name_newick_list):
    if not tree_name_newick_list:
        return ''
    f, wrapper = get_utf_8_string_io_writer()
    wrapper.write('''#NEXUS
BEGIN TAXA;
    Dimensions NTax = {s};
    TaxLabels {l} ;
END;
BEGIN TREES;
'''.format(s=len(quoted_leaf_labels), l=' '.join(quoted_leaf_labels)))
    for name, newick in tree_name_newick_list:
        wrapper.write('    Tree ')
        wrapper.write(name)
        wrapper.write(' = ')
        wrapper.write(newick)
    wrapper.write('\nEND;\n')
    flush_utf_8_writer(wrapper)
    return f.getvalue()


def convert_tree(tree_id, tree, otu_group, schema, subtree_id=None):
    label_key = schema.otu_label_prop
    if schema.format_str == 'nexus':
        leaf_labels = ([], {})
        needs_quotes_pattern = NEXUS_NEEDING_QUOTING
    else:
        leaf_labels = None
        needs_quotes_pattern = NEWICK_NEEDING_QUOTING
        assert schema.format_str == 'newick'
    newick = convert_tree_to_newick(tree,
                                    otu_group,
                                    label_key,
                                    leaf_labels,
                                    needs_quotes_pattern,
                                    subtree_id=subtree_id,
                                    bracket_ingroup=schema.bracket_ingroup)
    if schema.format_str == 'nexus':
        tl = [(quote_newick_name(tree_id, needs_quotes_pattern), newick)]
        return _write_nexus_format(leaf_labels[0], tl)
    else:
        return newick


def convert_trees(tid_tree_otus_list, schema, subtree_id=None):
    label_key = schema.otu_label_prop
    if schema.format_str == 'nexus':
        leaf_labels = ([], {})
        needs_quotes_pattern = NEXUS_NEEDING_QUOTING
        conv_tree_list = []
        for tree_id, tree, otu_group in tid_tree_otus_list:
            newick = convert_tree_to_newick(tree,
                                            otu_group,
                                            label_key,
                                            leaf_labels,
                                            needs_quotes_pattern,
                                            subtree_id=subtree_id,
                                            bracket_ingroup=schema.bracket_ingroup)
            if newick:
                t = (quote_newick_name(tree_id, needs_quotes_pattern), newick)
                conv_tree_list.append(t)
        return _write_nexus_format(leaf_labels[0], conv_tree_list)
    else:
        raise NotImplementedError('convert_tree for {}'.format(schema.format_str))


def nexml_el_of_by_id(nexson, curr_version=None):
    if curr_version is None:
        curr_version = detect_nexson_version(nexson)
    if not _is_by_id_hbf(curr_version):
        nexson = convert_nexson_format(nexson, BY_ID_HONEY_BADGERFISH)
    return get_nexml_el(nexson)


def extract_otus_nexson(nexson, otus_id, curr_version=None):
    nexml_el = nexml_el_of_by_id(nexson, curr_version)
    o = nexml_el['otusById']
    if otus_id is None:
        return o
    n = o.get(otus_id)
    if n is None:
        return None
    return {otus_id: n}


def extract_otu_nexson(nexson, otu_id, curr_version=None):
    nexml_el = nexml_el_of_by_id(nexson, curr_version)
    o = nexml_el['otusById']
    if otu_id is None:
        r = {}
        for g in o.values():
            r.update(g.get('otuById', {}))
        return r
    else:
        for g in o.values():
            go = g['otuById']
            if otu_id in go:
                return {otu_id: go[otu_id]}
    return None


def cull_nonmatching_trees(nexson, tree_id, curr_version=None):
    """Modifies `nexson` and returns it in version 1.2.1
    with any tree that does not match the ID removed.

    Note that this does not search through the NexSON for
    every node, edge, tree that was deleted. So the resulting
    NexSON may have broken references !
    """
    if curr_version is None:
        curr_version = detect_nexson_version(nexson)
    if not _is_by_id_hbf(curr_version):
        nexson = convert_nexson_format(nexson, BY_ID_HONEY_BADGERFISH)

    nexml_el = get_nexml_el(nexson)
    tree_groups = nexml_el['treesById']
    tree_groups_to_del = []
    for tgi, tree_group in tree_groups.items():
        tbi = tree_group['treeById']
        if tree_id in tbi:
            trees_to_del = [i for i in tbi.keys() if i != tree_id]
            for tid in trees_to_del:
                tree_group['^ot:treeElementOrder'].remove(tid)
                del tbi[tid]
        else:
            tree_groups_to_del.append(tgi)
    for tgid in tree_groups_to_del:
        nexml_el['^ot:treesElementOrder'].remove(tgid)
        del tree_groups[tgid]
    return nexson


def extract_tree_nexson(nexson, tree_id, curr_version=None):
    """Returns a list of (id, tree, otus_group) tuples for the
    specified tree_id (all trees if tree_id is None)
    """
    if curr_version is None:
        curr_version = detect_nexson_version(nexson)
    if not _is_by_id_hbf(curr_version):
        nexson = convert_nexson_format(nexson, BY_ID_HONEY_BADGERFISH)

    nexml_el = get_nexml_el(nexson)
    tree_groups = nexml_el['treesById']
    tree_obj_otus_group_list = []
    for tree_group in tree_groups.values():
        if tree_id:
            tree_list = [(tree_id, tree_group['treeById'].get(tree_id))]
        else:
            tree_list = tree_group['treeById'].items()
        for tid, tree in tree_list:
            if tree is not None:
                otu_groups = nexml_el['otusById']
                ogi = tree_group['@otus']
                otu_group = otu_groups[ogi]['otuById']
                tree_obj_otus_group_list.append((tid, tree, otu_group))
                if tree_id is not None:
                    return tree_obj_otus_group_list
    return tree_obj_otus_group_list


def extract_tree(nexson, tree_id, schema, subtree_id=None):
    try:
        assert schema.format_str in ['newick', 'nexus']
    except:
        m = 'Only newick tree export with tip labeling as one of "{}" is currently supported'
        m = m.format('", "'.join(PhyloSchema._NEWICK_PROP_VALS))
        raise ValueError(m)
    i_t_o_list = extract_tree_nexson(nexson, tree_id, None)
    if schema.format_str == 'newick':
        tree_str_list = [convert_tree(i, t, o, schema, subtree_id=subtree_id) for i, t, o in i_t_o_list]
        tree_str_list = [i for i in tree_str_list if i is not None]
        return '\n'.join(tree_str_list)
    return convert_trees(i_t_o_list, schema, subtree_id=subtree_id)


_DEF_MESSAGES_OBJ = {"message": tuple()}


def _get_supporting_file_messages_for_this_obj(o):
    m = []
    for i in o.get('^ot:messages', _DEF_MESSAGES_OBJ).get("message", []):
        if i.get("@code") == "SUPPORTING_FILE_INFO":
            m.append(i)
    return m


def extract_supporting_file_messages(nexson):
    curr_version = detect_nexson_version(nexson)
    if not _is_by_id_hbf(curr_version):
        nexson = convert_nexson_format(nexson, BY_ID_HONEY_BADGERFISH)
    nex = nexson['nexml']
    m_list = []
    m_list.extend(_get_supporting_file_messages_for_this_obj(nex))
    for otus in nex.get('otusById', {}).values():
        m_list.extend(_get_supporting_file_messages_for_this_obj(otus))
        for otu in otus.get('otuById', {}).values():
            m_list.extend(_get_supporting_file_messages_for_this_obj(otu))
    for tree_group in nex.get('treesById', {}).values():
        m_list.extend(_get_supporting_file_messages_for_this_obj(tree_group))
        for tree in tree_group.get('treeById', {}).values():
            m_list.extend(_get_supporting_file_messages_for_this_obj(tree))
    return m_list



#!/usr/bin/env python
import sys
import os

# setup.py largely based on
#   http://hynek.me/articles/sharing-your-labor-of-love-pypi-quick-and-dirty/
# Also see Jeet Sukumaran's DendroPy

###############################################################################
# setuptools/distutils/etc. import and configuration
try:
    # noinspection PyUnresolvedReferences
    import ez_setup

    try:
        ez_setup_path = " ('" + os.path.abspath(ez_setup.__file__) + "')"
    except OSError:
        ez_setup_path = ""
    sys.stderr.write("using ez_setup%s\n" % ez_setup_path)
    ez_setup.use_setuptools()
    # noinspection PyUnresolvedReferences
    import setuptools

    try:
        setuptools_path = " ('" + os.path.abspath(setuptools.__file__) + "')"
    except OSError:
        setuptools_path = ""
    sys.stderr.write("using setuptools%s\n" % setuptools_path)
    # noinspection PyUnresolvedReferences
    from setuptools import setup, find_packages
except ImportError as e:
    sys.stderr.write("using distutils\n")
    from distutils.core import setup

    sys.stderr.write("using canned package list\n")
    PACKAGES = ['nexson',
                'nexson.syntax',
                'nexson.validation',
                ]
    EXTRA_KWARGS = {}
else:
    sys.stderr.write("searching for packages\n")
    PACKAGES = find_packages()
    EXTRA_KWARGS = dict(
        include_package_data=True,
        test_suite="nexson.test"
    )

EXTRA_KWARGS["zip_safe"] = True
ENTRY_POINTS = {}
EXTRA_KWARGS['scripts'] = []

###############################################################################
# setuptools/distuils command extensions
try:
    from setuptools import Command
except ImportError:
    sys.stderr.write("setuptools.Command could not be imported: setuptools extensions not available\n")
else:
    sys.stderr.write("setuptools command extensions are available\n")
    command_hook = "distutils.commands"
    ENTRY_POINTS[command_hook] = []

setup(
    name='nexson',
    version='0.0.3',  # sync with __version__ in nexson/__init__.py
    description='Library for dealing with phylogenetic trees in the NexSON format',
    long_description=(open('README.rst').read()),
    url='https://github.com/OpenTreeOfLife/nexson',
    license='BSD',
    author='Emily Jane B. McTavish and Mark T. Holder',
    py_modules=['nexson'],
    install_requires=['setuptools', # sync with requirements.txt
                      'wheel',
                      'peyutil>=0.0.1' 
                      ],
    packages=PACKAGES,
    entry_points=ENTRY_POINTS,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    **EXTRA_KWARGS
)



#!/usr/bin/env python
"""Classes for different forms of Warnings and Errors.
"""


# An enum of WARNING_CODES
class NexsonWarningCodes(object):
    """Enumeration of Warning/Error types. For internal use.

    `facets` is a tuple mapping int -> warning name.
    Each of these names will also be an attribute of NexsonWarningCodes.
    NexsonWarningCodes.numeric_codes_registered is (after some mild monkey-patching)
        a set of the integers registered.
    """
    facets = ('MISSING_MANDATORY_KEY',
              'MISSING_OPTIONAL_KEY',
              'UNRECOGNIZED_KEY',
              'MISSING_LIST_EXPECTED',
              'DUPLICATING_SINGLETON_KEY',
              'REFERENCED_ID_NOT_FOUND',
              'REPEATED_ID',
              'REPEATED_OTU',
              'MULTIPLE_ROOT_NODES',
              'NO_ROOT_NODE',
              'MULTIPLE_EDGES_FOR_NODES',
              'CYCLE_DETECTED',
              'DISCONNECTED_GRAPH_DETECTED',
              'INCORRECT_ROOT_NODE_LABEL',
              'TIP_WITHOUT_OTU',
              'TIP_WITHOUT_OTT_ID',
              'MULTIPLE_TIPS_MAPPED_TO_OTT_ID',
              'NON_MONOPHYLETIC_TIPS_MAPPED_TO_OTT_ID',
              'INVALID_PROPERTY_VALUE',
              'PROPERTY_VALUE_NOT_USEFUL',
              'UNRECOGNIZED_PROPERTY_VALUE',
              'MULTIPLE_TREES',
              'UNVALIDATED_ANNOTATION',
              'UNRECOGNIZED_TAG',
              'CONFLICTING_PROPERTY_VALUES',
              'NO_TREES',
              'DEPRECATED_PROPERTY',
              'UNPARSEABLE_META',
              'UNREACHABLE_NODE',
              'INCORRECT_VALUE_TYPE',
              'MISSING_CRUCIAL_CONTENT',
              'MAX_SIZE_EXCEEDED',
              )
    numeric_codes_registered = []


# monkey-patching NexsonWarningCodes...
_mpdoc = ['    Code name -> int mapping:']
for _n, _f in enumerate(NexsonWarningCodes.facets):
    setattr(NexsonWarningCodes, _f, _n)
    _mpdoc.append('`{}` = {}'.format(_f, _n))
    NexsonWarningCodes.numeric_codes_registered.append(_n)
NexsonWarningCodes.numeric_codes_registered = set(NexsonWarningCodes.numeric_codes_registered)

NexsonWarningCodes.__doc__ = '{}\n{}\n'.format(NexsonWarningCodes.__doc__,
                                               '\n      - '.join(_mpdoc))
del _mpdoc
# End of NexsonWarningCodes enum

################################################################################
# In a burst of over-exuberant OO-coding, MTH added a class for
#   each class of Warning/Error.
#
# Each subclass typically tweaks the writing of the message and the payload
#   that constitutes the "data" blob in the JSON.
################################################################################
# class WarningMessage(object):
#     '''This base class provides the basic functionality of keeping
#     track of the "address" of the element that triggered the warning,
#     the severity code, and methods for writing to free text stream or JSON.
#     '''
#     def __init__(self,
#                  warning_code,
#                  data,
#                  address,
#                  severity=SeverityCodes.WARNING):
#         '''
#             `warning_code` should be a facet of NexsonWarningCodes
#             `data` is an object whose details depend on the specific subclass
#                 of warning that is being created
#             `address` is a NexsonAddress offending element

#             `severity` is either SeverityCodes.WARNING or SeverityCodes.ERROR
#         '''
#         self.warning_code = warning_code
#         assert warning_code in NexsonWarningCodes.numeric_codes_registered
#         self.warning_data = data
#         self.severity = severity
#         assert severity in SeverityCodes.numeric_codes_registered
#         self.address = address
#     def write(self, s, prefix):
#         raise NotImplementedError('WarningMessage.write')
#     def __unicode__(self, prefix=''):
#         b = StringIO()
#         ci = codecs.lookup('utf8')
#         s = codecs.StreamReaderWriter(b, ci.streamreader, ci.streamwriter)
#         self.write(s, prefix)
#         return s.getvalue()
#     def getvalue(self, prefix=''):
#         return self.__unicode__(prefix=prefix)
#     def as_dict(self):
#         return {
#             'severity': SeverityCodes.facets[self.severity],
#             'code': NexsonWarningCodes.facets[self.warning_code],
#             'comment': self.__unicode__(),
#             'data': self.convert_data_for_json(),
#             'refersTo': self.address.path
#         }
#     def convert_data_for_json(self):
#         data = self.warning_data
#         return data
#     def _write_message_suffix(self, out):
#         self.address.write_path_suffix_str(out)



#! /usr/bin/env python
from nexson import NexsonProxy
from nexson.test.support.pathmap import get_test_path_mapper
import unittest

pathmap = get_test_path_mapper()

class TestProxy(unittest.TestCase):
    def setUp(self):
        blob = pathmap.nexson_obj('9/v1.2.json')
        self.np = NexsonProxy(nexson=blob)

    def testGetters(self):
        self.assertIs(self.np.get_tree('bogus'), None)
        self.assertIsNot(self.np.get_tree('tree1'), None)
        self.assertIs(self.np.get_tree('otu123'), None)
        self.assertIs(self.np.get_otu('bogus'), None)
        self.assertIs(self.np.get_otu('tree1'), None)
        self.assertIsNot(self.np.get_otu('otu123'), None)

    def testCaching(self):
        f = self.np.get_tree('tree1')
        s = self.np.get_tree('tree1')
        self.assertIs(f, s)
        fn = f.get_node('node247')
        sn = s.get_node('node247')
        self.assertIs(fn, sn)
        fo = self.np.get_otu('otu123')
        so = sn.otu
        self.assertIs(fo, so)

    def testTreeIter(self):
        ntp = self.np.get_tree('tree1')
        nbi = ntp['nodeById']
        ebi = ntp['edgeBySourceId']
        ks = set(nbi.keys())
        its = set()
        edge_less = None
        for node in ntp:
            its.add(node.node_id)
            self.assertIs(node.node, nbi[node.node_id])
            if node.edge is None:
                self.assertIs(edge_less, None)
                edge_less = node.node_id
            else:
                s = node.edge['@source']
                ed = ebi[s]
                self.assertIs(node.edge, ed[node.edge_id])
        self.assertEqual(ntp['^ot:rootNodeId'], edge_less)
        self.assertEqual(ks, its)


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python
import unittest
import logging
from nexson.syntax import sort_arbitrarily_ordered_nexson
from nexson.manip import merge_otus_and_trees
from nexson.test.support.pathmap import get_test_path_mapper as ntest_path_mapper

pathmap = ntest_path_mapper()

_LOG = logging.getLogger(__name__)


class TestMerge(unittest.TestCase):
    def testCanConvert(self):
        inp = pathmap.nexson_obj('merge/merge-input.v1.2.json')
        expected = pathmap.nexson_obj('merge/merge-expected.v1.2.json')
        expected = sort_arbitrarily_ordered_nexson(expected)
        inp = sort_arbitrarily_ordered_nexson(inp)
        self.assertNotEqual(inp, expected)
        merge_otus_and_trees(inp)
        pathmap.equal_blob_check(self, '', inp, expected)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from nexson.syntax import (can_convert_nexson_forms,
                           convert_nexson_format,
                           DIRECT_HONEY_BADGERFISH,
                           BADGER_FISH_NEXSON_VERSION,
                           BY_ID_HONEY_BADGERFISH,
                           sort_meta_elements,
                           sort_arbitrarily_ordered_nexson)
from nexson.test import get_test_path_mapper
import unittest
import logging
import os

pathmap = get_test_path_mapper()

# pylint does not realize that serialize returns a string, so generates lots of
#   spurious errors
# pylint: disable=E1103
# round trip filename tuples
RT_DIRS = ['otu', '9', ]


def _get_pair(par, f, s):
    bf = os.path.join(par, f)
    hbf = os.path.join(par, s)
    fp, sp = pathmap.nexson_source_path(bf), pathmap.nexson_source_path(hbf)
    if not os.path.exists(fp):
        logging.getLogger(__name__).warning('\nTest skipped because {s} does not exist'.format(s=fp))
        return None, None
    if not os.path.exists(sp):
        logging.getLogger(__name__).warning('\nTest skipped because {s} does not exist'.format(s=sp))
        return None, None
    return pathmap.nexson_obj(bf), pathmap.nexson_obj(hbf)


class TestConvert(unittest.TestCase):
    def testCanConvert(self):
        x = ["0.0.0", "1.0.0"]
        for i in x:
            for j in x:
                self.assertTrue(can_convert_nexson_forms(i, j))

    def testCannotConvert(self):
        x = ["0.0.0", "1.1.0"]
        for i in x:
            for j in x:
                if i == j:
                    continue
                self.assertFalse(can_convert_nexson_forms(i, j))

    def testConvertBFtoHBF1_0(self):
        for t in RT_DIRS:
            obj, b_expect = _get_pair(t, 'v0.0.json', 'v1.0.json')
            if obj is None:
                continue
            h = convert_nexson_format(obj, DIRECT_HONEY_BADGERFISH)
            pathmap.equal_blob_check(self, '', h, b_expect)

    def testConvertBFtoHBF1_2(self):
        for t in RT_DIRS:
            obj, b_expect = _get_pair(t, 'v0.0.json', 'v1.2.json')
            if obj is None:
                continue
            b = convert_nexson_format(obj, BY_ID_HONEY_BADGERFISH)
            pathmap.equal_blob_check(self, '', b, b_expect)

    def testConvertHBF1_0toBF(self):
        for t in RT_DIRS:
            obj, b_expect = _get_pair(t, 'v1.0.json', 'v0.0.json')
            if obj is None:
                continue
            b = convert_nexson_format(obj, BADGER_FISH_NEXSON_VERSION)
            sort_meta_elements(b_expect)
            sort_meta_elements(b)
            pathmap.equal_blob_check(self, '', b, b_expect)

    def testConvertHBF1_2toBF(self):
        for t in RT_DIRS:
            obj, b_expect = _get_pair(t, 'v1.2.json', 'v0.0.json')
            if obj is None:
                continue
            b = convert_nexson_format(obj, BADGER_FISH_NEXSON_VERSION)
            sort_arbitrarily_ordered_nexson(b_expect)
            sort_arbitrarily_ordered_nexson(b)
            pathmap.equal_blob_check(self, '', b, b_expect)

    def testConvertHBF1_2toHBF1_0(self):
        for t in RT_DIRS:
            obj, b_expect = _get_pair(t, 'v1.2.json', 'v1.0.json')
            if obj is None:
                continue
            b = convert_nexson_format(obj, DIRECT_HONEY_BADGERFISH)
            sort_arbitrarily_ordered_nexson(b_expect)
            sort_arbitrarily_ordered_nexson(b)
            pathmap.equal_blob_check(self, '', b, b_expect)

    def testConvertHBF1_0toHBF1_2(self):
        for t in RT_DIRS:
            obj, b_expect = _get_pair(t, 'v1.0.json', 'v1.2.json')
            if obj is None:
                continue
            b = convert_nexson_format(obj, BY_ID_HONEY_BADGERFISH)
            pathmap.equal_blob_check(self, '', b, b_expect)


if __name__ == "__main__":
    unittest.main()
